# Copyright 2020 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=no-else-return
# pylint: disable=duplicate-code

import unittest

#from okera import context, _thrift_api
#from datetime import datetime
from okera.tests import pycerebro_test_common as common
from okera._thrift_api import (
    TDataRegConnection, TRecordServiceException, TCrawlStatus,
    TAccessPermissionLevel, TDataRegConnection)

from okera import _thrift_api

EMPTY_DRC = TDataRegConnection()
EMPTY_DRC.name = ''

# level, can show, can alter, can drop, can use
CONNECTION_PRIVILEGES = [
    (TAccessPermissionLevel.ALL, True, True, True, True),
    (TAccessPermissionLevel.SHOW, True, False, False, False),
    (TAccessPermissionLevel.ALTER, True, True, False, False),
    (TAccessPermissionLevel.DROP, True, False, True, False),
    (TAccessPermissionLevel.USE, True, False, False, True),
]

def get_connection_as_root(conn, name):
    drc = TDataRegConnection()
    drc.name = name

    orig_user = conn.ctx._get_user()
    conn.ctx.disable_auth()
    connections = conn.manage_data_reg_connection("GET", drc).connections
    conn.ctx.enable_token_auth(token_str=orig_user)

    return connections[0]

def get_connection(conn, name):
    drc = TDataRegConnection()
    drc.name = name

    connections = conn.manage_data_reg_connection("GET", drc).connections
    return connections[0]

def delete_connection(conn, name):
    drc = get_connection_as_root(conn, name)

    return conn.manage_data_reg_connection("DELETE", drc)

def update_connection(conn, name):
    drc = get_connection_as_root(conn, name)

    return conn.manage_data_reg_connection("UPDATE", drc)

def create_connection(conn, drc):
    return conn.manage_data_reg_connection("CREATE", drc)

def run_test_connection(conn, name):
    drc = get_connection_as_root(conn, name)

    return conn.manage_data_reg_connection("TEST_EXISTING", drc)

# Run TEST on a connection prior to it being created
def run_test_connection_on_new(conn, drc):
    return conn.manage_data_reg_connection("TEST_CREATE", drc)

def discover_crawler(conn, name):
    drc = get_connection_as_root(conn, name)

    return conn.discover_crawler(drc)

def manage_crawler(conn, name):
    drc = get_connection_as_root(conn, name)

    return conn.manage_crawler(drc)

class DataRegistrationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """ Initializes one time state that is shared across test cases. This is used
            to speed up the tests. State that can be shared across (but still stable)
            should be here instead of __cleanup()."""
        super(DataRegistrationTest, cls).setUpClass()
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS test_connection_name')
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS test_connection_name_2')
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS mysql_test_connection')
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS mssql_test_connection')
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS snwflk_test_connection')

    def _get_psqltest_data_reg_connection_obj(self, name):
        data_reg_connection = TDataRegConnection()
        data_reg_connection.name = name
        data_reg_connection.type = "JDBC"
        data_reg_connection.data_source_path = 's3://cerebro-datasets/' \
            'jdbc_demo/jdbc_test_psql.conf'
        data_reg_connection.jdbc_driver = "postgresql"
        data_reg_connection.host = "jdbcpsqltest.cyn8yfvyuugz.us-west-2.rds.amazonaws.com"
        data_reg_connection.port = 5432
        data_reg_connection.user_key = "awssm://postgres-jdbcpsqltest-username"
        data_reg_connection.password_key = "awssm://postgres-jdbcpsqltest-password"
        data_reg_connection.default_catalog = "jdbc_test"
        data_reg_connection.default_schema = "public"
        data_reg_connection.is_active = True
        data_reg_connection.connection_properties = {'ssl':'False'}
        return data_reg_connection

    def _get_mysqltest_data_reg_connection_obj(self, name):
        data_reg_connection = TDataRegConnection()
        data_reg_connection.name = name
        data_reg_connection.type = "JDBC"
        data_reg_connection.data_source_path = None
        data_reg_connection.jdbc_driver = "mysql"
        data_reg_connection.host = 'cerebro-db-test-long-running.cyn8yfvyuugz.' \
            'us-west-2.rds.amazonaws.com'
        data_reg_connection.port = 3306
        data_reg_connection.user_key = "awssm://mysql-username"
        data_reg_connection.password_key = "awssm://mysql-password"
        data_reg_connection.default_catalog = "jdbc_test"
        data_reg_connection.default_schema = None
        data_reg_connection.is_active = True
        data_reg_connection.connection_properties = {'ssl':'False'}
        return data_reg_connection

    def _get_mssqltest_data_reg_connection_obj(self, name):
        data_reg_connection = TDataRegConnection()
        data_reg_connection.name = name
        data_reg_connection.type = "JDBC"
        data_reg_connection.data_source_path = None
        data_reg_connection.jdbc_driver = "sqlserver"
        data_reg_connection.host = 'mssql-server-test.cyn8yfvyuugz.' \
            'us-west-2.rds.amazonaws.com'
        data_reg_connection.port = 1433
        data_reg_connection.user_key = "awssm://mssql-username"
        data_reg_connection.password_key = "awssm://mssql-password"
        data_reg_connection.default_catalog = "okera_test"
        data_reg_connection.default_schema = None
        data_reg_connection.is_active = True
        data_reg_connection.connection_properties = {'ssl':'False'}
        return data_reg_connection

    def _get_snowflake_data_reg_connection_obj(self, name):
        data_reg_connection = TDataRegConnection()
        data_reg_connection.name = name
        data_reg_connection.type = "JDBC"
        data_reg_connection.data_source_path = None
        data_reg_connection.jdbc_driver = "snowflake"
        data_reg_connection.host = "vq85960.snowflakecomputing.com"
        data_reg_connection.port = 0
        data_reg_connection.user_key = "awssm://snowflake-partner-username"
        data_reg_connection.password_key = "awssm://snowflake-partner-password"
        data_reg_connection.default_catalog = "SNOWFLAKE_SAMPLE_DATA"
        data_reg_connection.default_schema = "TPCH_SF1"
        data_reg_connection.is_active = True
        data_reg_connection.connection_properties = {'defaultDb':'TPCH_SF1'}
        return data_reg_connection

    def _verify_mssql_table_schema(self, ret_list_jdbc_datasets, \
        ret_jdbc_schema, ret_jdbc_table):
        for dataset in ret_list_jdbc_datasets:
            if dataset.jdbc_schema == ret_jdbc_schema and \
                dataset.jdbc_table == ret_jdbc_table:
                for col in dataset.schema.cols:
                    if col.name == 'varchar':
                        self.assertTrue(col.type.type_id == 8) #TTypeId=8=VARCHAR
                        self.assertTrue(col.type.len == 20)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'text':
                        self.assertTrue(col.type.type_id == 7)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'tinyint':
                        self.assertTrue(col.type.type_id == 1)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'smallint':
                        self.assertTrue(col.type.type_id == 2)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'int':
                        self.assertTrue(col.type.type_id == 3)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'bigint':
                        self.assertTrue(col.type.type_id == 4)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'date':
                        self.assertTrue(col.type.type_id == 16)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'float_col':
                        self.assertTrue(col.type.type_id == 6)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'decimal':
                        self.assertTrue(col.type.type_id == 10)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision == 10)
                        self.assertTrue(col.type.scale == 2)
                    elif col.name == 'datetime':
                        self.assertTrue(col.type.type_id == 11)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'timestamp':
                        self.assertTrue(col.type.type_id == 15)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'time':
                        self.assertTrue(col.type.type_id == 11)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'char':
                        self.assertTrue(col.type.type_id == 9)
                        self.assertTrue(col.type.len == 10)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'binary':
                        self.assertTrue(col.type.type_id == 15)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'varbinary':
                        self.assertTrue(col.type.type_id == 15)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)

    def _verify_mysql_table_schema(self, ret_list_jdbc_datasets, \
        ret_jdbc_schema, ret_jdbc_table):
        for dataset in ret_list_jdbc_datasets:
            if dataset.jdbc_schema == ret_jdbc_schema and \
                dataset.jdbc_table == ret_jdbc_table:
                for col in dataset.schema.cols:
                    if col.name == 'varchar':
                        self.assertTrue(col.type.type_id == 8) #TTypeId=8=VARCHAR
                        self.assertTrue(col.type.len == 20)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'text':
                        self.assertTrue(col.type.type_id == 8)
                        self.assertTrue(col.type.len == 65355)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'double':
                        self.assertTrue(col.type.type_id == 6)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'date':
                        self.assertTrue(col.type.type_id == 16)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'float':
                        self.assertTrue(col.type.type_id == 6)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'decimal':
                        self.assertTrue(col.type.type_id == 10)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision == 10)
                        self.assertTrue(col.type.scale == 2)
                    elif col.name == 'datetime':
                        self.assertTrue(col.type.type_id == 11)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'timestamp':
                        self.assertTrue(col.type.type_id == 11)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'time':
                        self.assertTrue(col.type.type_id == 11)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'year':
                        self.assertTrue(col.type.type_id == 16)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'char':
                        self.assertTrue(col.type.type_id == 9)
                        self.assertTrue(col.type.len == 10)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'enum':
                        self.assertTrue(col.type.type_id == 9)
                        self.assertTrue(col.type.len == 1)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'set':
                        self.assertTrue(col.type.type_id == 9)
                        self.assertTrue(col.type.len == 5)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)
                    elif col.name == 'bool':
                        self.assertTrue(col.type.type_id == 0)
                        self.assertTrue(col.type.len is None)
                        self.assertTrue(col.type.precision is None)
                        self.assertTrue(col.type.scale is None)

    def test_drc_crud_backwards_compat_test(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:

            # Clean if already exists
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS test_connection_name_bc')
            drc = self._get_psqltest_data_reg_connection_obj("test_connection_name_bc")

            # create test
            drc.username = "awssm://postgres-jdbcpsqltest-username"
            drc.password = "awssm://postgres-jdbcpsqltest-password"
            drcs = conn.manage_data_reg_connection("CREATE", drc)
            self.assertTrue(len(drcs.connections) == 1)

            # update test
            drc.host = "test_host_name_mod_bc"
            drcs = conn.manage_data_reg_connection("UPDATE", drc)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertEqual(drcs.connections[0].host,
                             "test_host_name_mod_bc",
                             "Update data registration object connection failed.")

            # get test
            drcs = conn.manage_data_reg_connection("GET", drc)
            self.assertTrue(len(drcs.connections) == 1)

            # list with filters works as expected
            drcs = conn.manage_data_reg_connection(
                "LIST", drc, ["test_connection_name_bc"])
            self.assertTrue(len(drcs.connections) >= 1)
            connection_names = [names.name for names in drcs.connections]
            self.assertTrue("test_connection_name_bc" in connection_names,
                            "list data registration object connection failed.")

            # create another one
            drc = self._get_psqltest_data_reg_connection_obj("test_connection_name_bc_2")
            drcs = conn.manage_data_reg_connection("CREATE", drc)
            self.assertTrue(len(drcs.connections) == 1)

            # list with empty filter should still return all
            drcs = conn.manage_data_reg_connection(
                "LIST", drc, [])
            self.assertTrue(len(drcs.connections) >= 2)
            connection_names = [names.name for names in drcs.connections]
            self.assertTrue("test_connection_name_bc" in connection_names,
                            "list data registration object connection failed.")
            self.assertTrue("test_connection_name_bc_2" in connection_names,
                            "list data registration object connection failed.")

            # list with pattern search works as expected
            drcs = conn.manage_data_reg_connection(
                "LIST", drc, ["test_connection_name_bc", "test_connection_name_bc_2"])
            self.assertTrue(len(drcs.connections) >= 2)

            # delete test
            drcs = conn.manage_data_reg_connection("DELETE", drc)
            self.assertTrue(len(drcs.connections) == 0)

            # get again returns empty
            drcs = conn.manage_data_reg_connection("GET", drc)
            self.assertTrue(len(drcs.connections) == 0)

    def test_drc_crud(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # create test
            drc = self._get_psqltest_data_reg_connection_obj("test_connection_name")
            drcs = conn.manage_data_reg_connection("CREATE", drc)
            self.assertTrue(len(drcs.connections) == 1)

            # update test
            drc.host = "test_host_name_mod"
            drcs = conn.manage_data_reg_connection("UPDATE", drc)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertEqual(drcs.connections[0].host,
                             "test_host_name_mod",
                             "Update data registration object connection failed.")

            # get test
            drcs = conn.manage_data_reg_connection("GET", drc)
            self.assertTrue(len(drcs.connections) == 1)

            # list with filters works as expected
            drcs = conn.manage_data_reg_connection(
                "LIST", drc, ["test_connection_name"])
            self.assertTrue(len(drcs.connections) >= 1)
            connection_names = [names.name for names in drcs.connections]
            self.assertTrue("test_connection_name" in connection_names,
                            "list data registration object connection failed.")

            # create another one
            drc = self._get_psqltest_data_reg_connection_obj("test_connection_name_2")
            drcs = conn.manage_data_reg_connection("CREATE", drc)
            self.assertTrue(len(drcs.connections) == 1)

            # list with empty filter should still return all
            drcs = conn.manage_data_reg_connection(
                "LIST", drc, [])
            self.assertTrue(len(drcs.connections) >= 2)
            connection_names = [names.name for names in drcs.connections]
            self.assertTrue("test_connection_name" in connection_names,
                            "list data registration object connection failed.")
            self.assertTrue("test_connection_name_2" in connection_names,
                            "list data registration object connection failed.")

            # list with pattern search works as expected
            drcs = conn.manage_data_reg_connection(
                "LIST", drc, ["test_connection_name", "test_connection_name_2"])
            self.assertTrue(len(drcs.connections) >= 2)
            ## Fixme: This is broken, looks like was always broken
            # self.assertEqual(drcs.connections[0].name,
            #                  "test_connection_name",
            #                  "list data registration object connection failed.")
            # self.assertEqual(drcs.connections[1].name,
            #                  "test_connection_name_2",
            #                  "list data registration object connection failed.")

            # Fixme:: Not yet working, returns all
            # list with pattern search works as expected
            # drcs = conn.manage_data_reg_connection(
            #     "LIST", drc, [], "test_connection_name_2")
            # print(drcs.connections)
            # self.assertTrue(len(drcs.connections) == 1)
            # self.assertEqual(drcs.connections[0].name,
            #                  "test_connection_name_2",
            #                  "list data registration object connection failed." +
            #                  len(drcs.connections))

            # delete test
            drcs = conn.manage_data_reg_connection("DELETE", drc)
            self.assertTrue(len(drcs.connections) == 0)

            # get again returns empty
            drcs = conn.manage_data_reg_connection("GET", drc)
            self.assertTrue(len(drcs.connections) == 0)

    def test_drc_testconnection_as_admin(self):
        MYSQL_CXN = 'test_drc_testconnection_as_admin_mysql_conn'
        SNWFLK_CXN = 'test_drc_testconnection_as_admin_snwflk_conn'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()
            ddls = [
                "DROP DATACONNECTION IF EXISTS %s" % (MYSQL_CXN),
                "DROP DATACONNECTION IF EXISTS %s" % (SNWFLK_CXN),
            ]
            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Test Connection in "CREATE WORKFLOW" with valid connection details, TEST should succeed
            drc = self._get_mysqltest_data_reg_connection_obj(MYSQL_CXN)
            drcs = conn.manage_data_reg_connection("TEST_CREATE", drc)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].test_status == "SUCCESS")
            self.assertTrue(drcs.connections[0].test_error_details == "")
            self.assertTrue(drcs.connections[0].tested_at is not None)
            # The connection should not get created as it is just a TEST
            drcs = conn.manage_data_reg_connection("GET", drc)
            self.assertTrue(len(drcs.connections) == 0)

            # Test Connection in "CREATE WORKFLOW" with missing default db, TEST should fail
            drc = self._get_mysqltest_data_reg_connection_obj(MYSQL_CXN)
            # set the default catalog to be Null
            drc.default_catalog = None
            drcs = conn.manage_data_reg_connection("TEST_CREATE", drc)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].test_status == "FAILED")
            self.assertEqual(drcs.connections[0].test_error_details,
                "'jdbc.db.name' is a required parameter for a data connection.")
            self.assertTrue(drcs.connections[0].tested_at is not None)
            # The connection should not get created as it is just a TEST
            drcs = conn.manage_data_reg_connection("GET", drc)
            self.assertTrue(len(drcs.connections) == 0)

            # Test Connection in "EDIT WORKFLOW" with invalid connection details, TEST should fail
            drc = self._get_mysqltest_data_reg_connection_obj(MYSQL_CXN)
            # create a connection
            drcs1 = conn.manage_data_reg_connection("CREATE", drc)
            self.assertTrue(len(drcs1.connections) == 1)
            # Now edit some connection details and test the updated connection
            drc.host = "invalid_mysql_host"
            drcs2 = conn.manage_data_reg_connection("TEST_EDIT", drc)
            self.assertTrue(len(drcs2.connections) == 1)
            self.assertTrue(drcs2.connections[0].test_status == "FAILED")
            self.assertEqual(drcs2.connections[0].test_error_details,
                "com.cloudera.impala.common.AnalysisException: JDBC data source " \
                "connection validation failed. Please check connection properties." \
                "\nError from 'mysql' data source.")
            self.assertTrue(drcs2.connections[0].tested_at is not None)

            # Test Connection in "EDIT WORKFLOW" with invalid connection details, TEST should fail
            drc = self._get_snowflake_data_reg_connection_obj(SNWFLK_CXN)
            # create a connection
            drcs1 = conn.manage_data_reg_connection("CREATE", drc)
            self.assertTrue(len(drcs1.connections) == 1)
            # Now edit the userkey to be mysql's userkey and test the updated connection
            drc.user_key = "awssm://mysql-username"
            drcs2 = conn.manage_data_reg_connection("TEST_EDIT", drc)
            self.assertTrue(len(drcs2.connections) == 1)
            self.assertTrue(drcs2.connections[0].test_status == "FAILED")
            self.assertEqual(drcs2.connections[0].test_error_details,
                "com.cloudera.impala.common.AnalysisException: JDBC data source " \
                "connection validation failed. Please check connection properties." \
                "\nError from 'snowflake' data source.")
            self.assertTrue(drcs2.connections[0].tested_at is not None)

            # "TEST EXISTING" connection with valid connection details, TEST should succeed
            drc = self._get_mysqltest_data_reg_connection_obj(MYSQL_CXN)
            # drop connection if exists
            drcs1 = conn.manage_data_reg_connection("DELETE", drc)
            # create a connection
            drcs1 = conn.manage_data_reg_connection("CREATE", drc)
            self.assertTrue(len(drcs1.connections) == 1)
            # Now test an existing connection
            drcs2 = conn.manage_data_reg_connection("TEST_EXISTING", drc)
            self.assertTrue(len(drcs2.connections) == 1)
            self.assertTrue(drcs2.connections[0].test_status == "SUCCESS")
            self.assertTrue(drcs2.connections[0].test_error_details == "")
            self.assertTrue(drcs2.connections[0].tested_at is not None)
            # The existing connection should get updated with latest test information
            drcs3 = conn.manage_data_reg_connection("GET", drc)
            self.assertTrue(len(drcs3.connections) == 1)
            self.assertTrue(drcs3.connections[0].test_status == "SUCCESS")
            self.assertTrue(drcs3.connections[0].test_error_details == "")
            self.assertTrue(drcs3.connections[0].tested_at is not None)
            # the 'tested_at' timestamp should match
            # Note that we round the milliseconds away as some DB engines
            # (such as MySQL) do the same thing, and we just care that
            # overall they are similar.
            self.assertEqual(round(drcs2.connections[0].tested_at / 1000),
                round(drcs3.connections[0].tested_at / 1000))

    def test_invalid_jdbc_props(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # validation for invalid path
            drc = self._get_psqltest_data_reg_connection_obj('test_conn_invalid_path')
            drc.data_source_path = 's3://some-invalid-path/test.conf'
            try:
                drcs = conn.manage_data_reg_connection('CREATE', drc)
                assert drcs
            except TRecordServiceException as ex:
                assert 'JDBC credentials validation failed for data registration ' \
                    'connection with name ' + drc.name in str(ex.detail)

            drc = self._get_psqltest_data_reg_connection_obj('test_conn_invalid_path')
            drcs = conn.manage_data_reg_connection('CREATE', drc)
            self.assertTrue(len(drcs.connections) == 1)
            # validation for invalid path update
            drc.data_source_path = 's3://some-invalid-path/test.conf'
            try:
                drcs = conn.manage_data_reg_connection('UPDATE', drc)
                assert drcs
            except TRecordServiceException as ex:
                assert 'JDBC credentials validation failed for data registration ' \
                    'connection with name ' + drc.name in str(ex.detail)

            drcs = conn.manage_data_reg_connection('DELETE', drc)
            self.assertTrue(len(drcs.connections) == 0)

    def test_discover_crawler_catalog(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            #MySQL Test
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS mysql_test_connection')
            drc = self._get_mysqltest_data_reg_connection_obj('mysql_test_connection')
            drcs = conn.manage_data_reg_connection('CREATE', drc)
            ret_crawler_objects = conn.discover_crawler(drc)
            ret_list_jdbc_datasets = ret_crawler_objects.crawler_discover_datasets[0] \
                .jdbc_datasets
            self.assertTrue(len(ret_crawler_objects.crawler_discover_datasets) == 1)
            self.assertIsNotNone(ret_list_jdbc_datasets)
            self.assertTrue(any(x.jdbc_catalog == 'jdbc_test'\
                for x in ret_list_jdbc_datasets))

            #Microsoft SQL Server Test
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS mssql_test_connection')
            drc = self._get_mssqltest_data_reg_connection_obj('mssql_test_connection')
            drcs = conn.manage_data_reg_connection('CREATE', drc)
            ret_crawler_objects = conn.discover_crawler(drc)
            ret_list_jdbc_datasets = ret_crawler_objects.crawler_discover_datasets[0] \
                .jdbc_datasets

            self.assertTrue(len(ret_crawler_objects.crawler_discover_datasets) == 1)
            self.assertIsNotNone(ret_list_jdbc_datasets)
            self.assertTrue(any(x.jdbc_catalog == 'okera_test'\
                for x in ret_list_jdbc_datasets))

    def test_crawler(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            drc = self._get_psqltest_data_reg_connection_obj('test_conn_crawler')
            drc.data_source_path = 's3://cerebro-datasets/jdbc_demo/jdbc_test_psql.conf'
            # delete if it exists then create
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS test_conn_crawler')
            drcs = conn.manage_data_reg_connection("CREATE", drc)
            resp = conn.manage_crawler(drc)
            self.assertTrue(resp.status == TCrawlStatus.CRAWLING)

            drc = self._get_psqltest_data_reg_connection_obj('test_conn_crawler_s3')
            drc.type = "S3"
            drc.data_source_path = 's3://cerebrodata-test/tpch-nation/'
            resp = conn.manage_crawler(drc)
            self.assertTrue(resp.status == TCrawlStatus.CRAWLING)

    def test_discover_crawler_schema(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS mssql_test_connection')
            drc = self._get_mssqltest_data_reg_connection_obj('mssql_test_connection')
            drcs = conn.manage_data_reg_connection('CREATE', drc)
            ret_crawler_objects = conn.discover_crawler(drc, 'okera_test')
            ret_list_jdbc_datasets = ret_crawler_objects.crawler_discover_datasets[0] \
                .jdbc_datasets

            self.assertTrue(len(ret_crawler_objects.crawler_discover_datasets) == 1)
            self.assertIsNotNone(ret_list_jdbc_datasets)
            self.assertTrue(any(x.jdbc_catalog == 'okera_test' \
                for x in ret_list_jdbc_datasets))
            #for dataset in ret_list_jdbc_datasets:
            #    print(dataset.jdbc_schema)
            self.assertTrue(any(x.jdbc_schema == 'marketing' \
                for x in ret_list_jdbc_datasets))
            self.assertTrue(any(x.jdbc_schema == 'dbo' \
                for x in ret_list_jdbc_datasets))
            self.assertFalse(any(x.jdbc_schema == 'INFORMATION_SCHEMA' \
                for x in ret_list_jdbc_datasets))
            drcs = conn.manage_data_reg_connection('DELETE', drc)
            self.assertTrue(len(drcs.connections) == 0)

    def test_discover_crawler_table_schema(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            #MySQL Test
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS mysql_test_connection')
            drc = self._get_mysqltest_data_reg_connection_obj('mysql_test_connection')
            drcs = conn.manage_data_reg_connection('CREATE', drc)
            ret_crawler_objects = conn.discover_crawler(drc, 'jdbc_test')
            ret_list_jdbc_datasets = ret_crawler_objects.crawler_discover_datasets[0] \
                .jdbc_datasets

            self.assertTrue(len(ret_crawler_objects.crawler_discover_datasets) == 1)
            self.assertIsNotNone(ret_list_jdbc_datasets)
            self.assertTrue(any(x.jdbc_catalog == 'jdbc_test'\
                for x in ret_list_jdbc_datasets))
            #for dataset in ret_list_jdbc_datasets:
            #    print(dataset)
            self.assertTrue(any(x.jdbc_table == 'all_data_types' \
                for x in ret_list_jdbc_datasets))
            self._verify_mysql_table_schema(ret_list_jdbc_datasets, \
                None, 'all_data_types')
            drcs = conn.manage_data_reg_connection('DELETE', drc)
            self.assertTrue(len(drcs.connections) == 0)

            #Microsoft SQL Server Test
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS mssql_test_connection')
            drc = self._get_mssqltest_data_reg_connection_obj('mssql_test_connection')
            drcs = conn.manage_data_reg_connection('CREATE', drc)
            ret_crawler_objects = conn.discover_crawler(drc, 'okera_test',\
                 'okera_test.dbo')
            ret_list_jdbc_datasets = ret_crawler_objects.crawler_discover_datasets[0]\
                .jdbc_datasets

            self.assertTrue(len(ret_crawler_objects.crawler_discover_datasets) == 1)
            self.assertIsNotNone(ret_list_jdbc_datasets)
            self.assertTrue(any(x.jdbc_catalog == 'okera_test' \
                for x in ret_list_jdbc_datasets))
            self.assertTrue(any(x.jdbc_schema == 'okera_test.dbo' \
                for x in ret_list_jdbc_datasets))
            #for dataset in ret_list_jdbc_datasets:
            #    print(dataset.jdbc_table)
            self.assertTrue(any(x.jdbc_table == 'all_data_types' \
                for x in ret_list_jdbc_datasets))

            for dataset in ret_list_jdbc_datasets:
                if dataset.jdbc_schema == 'okera_test.dbo' and \
                   dataset.jdbc_table == 'dbo.all_data_types':
                    self.assertTrue(any(x.name == 'varchar' \
                        for x in dataset.schema.cols))
            self._verify_mssql_table_schema(ret_list_jdbc_datasets, \
                'okera_test.dbo', 'dbo.all_data_types')
            #for dataset in ret_list_jdbc_datasets:
            #    print(dataset)

            drcs = conn.manage_data_reg_connection('DELETE', drc)
            self.assertTrue(len(drcs.connections) == 0)

    def test_dataconnection_privileges(self):
        CXN = 'test_privileges_connection_privileges'
        CXN2 = 'test_privileges_connection_privileges_2'
        ROLE = "test_privileges_connection_role"

        USER = "test_privileges_connection_user"
        USER_NO_ACCESS = "test_privileges_connection_user_no_access"

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            for is_catalog in [False, True]:
                for level_test in CONNECTION_PRIVILEGES:
                    for is_ddl in [True, False]:
                        ctx.disable_auth()

                        level, can_show, can_alter, can_drop, can_use = level_test
                        level_name = TAccessPermissionLevel._VALUES_TO_NAMES[level]
                        ddls = [
                            "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                            "DROP DATACONNECTION IF EXISTS %s" % (CXN2),
                            "DROP ROLE IF EXISTS %s" % (ROLE),

                            # Create the connection
                            """CREATE DATACONNECTION %s CXNPROPERTIES
                                (
                                'connection_type'='JDBC',
                                'jdbc_driver'='mysql',
                                'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                                'port'='3306',
                                'user_key'='awsps:///mysql/username',
                                'password_key'='awsps:///mysql/password',
                                'jdbc.db.name'='jdbc_test'
                                )
                            """ % (CXN),

                            # Create the connection
                            """CREATE DATACONNECTION %s CXNPROPERTIES
                                (
                                'connection_type'='JDBC',
                                'jdbc_driver'='mysql',
                                'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                                'port'='3306',
                                'user_key'='awsps:///mysql/username',
                                'password_key'='awsps:///mysql/password',
                                'jdbc.db.name'='jdbc_test'
                                )
                            """ % (CXN2),

                            # Create the role and grant it to the user
                            "CREATE ROLE %s" % (ROLE),
                            "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),
                        ]


                        if is_catalog:
                            # TODO: CATALOG doesn't support ALTER
                            if level == TAccessPermissionLevel.ALTER:
                                continue

                            ddls += [
                                "GRANT %s ON CATALOG TO ROLE %s" % (level_name, ROLE)
                            ]
                        else:
                            ddls += [
                                "GRANT %s ON DATACONNECTION %s TO ROLE %s" % (level_name, CXN, ROLE)
                            ]

                        for ddl in ddls:
                            conn.execute_ddl(ddl)

                        # Verify that the root user can see both connections
                        ctx.disable_auth()
                        connections = conn.manage_data_reg_connection("LIST", EMPTY_DRC, connection_pattern=CXN).connections
                        assert len(connections) == 2
                        sorted_names = sorted([connection.name for connection in connections])
                        assert sorted_names[0] == CXN
                        assert sorted_names[1] == CXN2

                        ctx.enable_token_auth(token_str=USER)

                        connections = conn.manage_data_reg_connection("LIST", EMPTY_DRC, connection_pattern=CXN).connections
                        if can_show:
                            # Check to ensure the connection is visible. Note that if the permission we have
                            # is on the catalog level, we will be able to see both connections, so check
                            # for that explicitly.
                            if is_catalog:
                                assert len(connections) == 2
                                sorted_names = sorted([connection.name for connection in connections])
                                assert sorted_names[0] == CXN
                                assert sorted_names[1] == CXN2
                            else:
                                assert len(connections) == 1
                                assert connections[0].name == CXN
                        else:
                            assert len(connections) == 0

                        if is_ddl:
                            stmt = """ALTER DATACONNECTION %s CXNPROPERTIES
                                        (
                                        'connection_type'='JDBC',
                                        'jdbc_driver'='mysql',
                                        'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                                        'port'='3306',
                                        'user_key'='awsps:///mysql/username',
                                        'password_key'='awsps:///mysql/password',
                                        'jdbc.db.name'='jdbc_test',
                                        'foo'='bar'
                                        )""" % (CXN)
                            if can_alter:
                                conn.execute_ddl(stmt)
                            else:
                                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                    conn.execute_ddl(stmt)
                                self.assertTrue('ALTER this data connection' in str(ex_ctx.exception))
                        else:
                            if can_alter:
                                update_connection(conn, CXN)
                            else:
                                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                    update_connection(conn, CXN)
                                self.assertTrue('ALTER this data connection' in str(ex_ctx.exception))

                        # USE should give us the ability to TEST a connection
                        if is_ddl:
                            # Only test it via API, there is no TEST DDL for connections
                            pass
                        else:
                            if can_use:
                                run_test_connection(conn, CXN)
                            else:
                                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                    run_test_connection(conn, CXN)
                                self.assertTrue('test this data connection' in str(ex_ctx.exception))

                        if is_ddl:
                            # Only test it via API, there is no SHOW DDL for connections
                            pass
                        else:
                            if can_show:
                                get_connection(conn, CXN)
                            else:
                                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                    get_connection(conn, CXN)
                                self.assertTrue('SHOW this data connection' in str(ex_ctx.exception))

                        if is_ddl:
                            stmt = "DROP DATACONNECTION IF EXISTS %s" % (CXN)
                            if can_drop:
                                conn.execute_ddl(stmt)
                            else:
                                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                    conn.execute_ddl(stmt)
                                self.assertTrue('DROP this data connection' in str(ex_ctx.exception))
                        else:
                            if can_drop:
                                delete_connection(conn, CXN)
                            else:
                                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                    delete_connection(conn, CXN)
                                self.assertTrue('DROP this data connection' in str(ex_ctx.exception))


                        # Verify that a user with no access can't see the connection
                        ctx.enable_token_auth(token_str=USER_NO_ACCESS)
                        connections = conn.manage_data_reg_connection("LIST", EMPTY_DRC, connection_pattern=CXN).connections
                        assert len(connections) == 0

    def test_dataconnection_create_as_owner(self):
        CXN = 'test_privileges_connection_create_as_owner'
        ROLE = "test_privileges_connection_role"

        USER = "test_privileges_connection_user"
        USER_NO_ACCESS = "test_privileges_connection_user_no_access"

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            for is_ddl in [True, False]:
                ctx.disable_auth()

                ddls = [
                    "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                    "DROP ROLE IF EXISTS %s" % (ROLE),

                    # Create the role and grant it to the user
                    "CREATE ROLE %s" % (ROLE),
                    "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),

                    "GRANT CREATE_DATACONNECTION_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE),
                ]

                for ddl in ddls:
                    conn.execute_ddl(ddl)

                if is_ddl:
                    create_stmt = """CREATE DATACONNECTION %s CXNPROPERTIES
                        (
                        'connection_type'='JDBC',
                        'jdbc_driver'='mysql',
                        'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                        'port'='3306',
                        'user_key'='awsps:///mysql/username',
                        'password_key'='awsps:///mysql/password',
                        'jdbc.db.name'='jdbc_test'
                        )
                    """ % (CXN)
                    for user, can_create in [(USER, True), (USER_NO_ACCESS, False)]:
                        ctx.enable_token_auth(token_str=user)
                        if can_create:
                            conn.execute_ddl(create_stmt)

                            # Since we did create as owner, we can drop it
                            stmt = "DROP DATACONNECTION IF EXISTS %s" % (CXN)
                            conn.execute_ddl(stmt)
                        else:
                            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                conn.execute_ddl(create_stmt)
                            self.assertTrue('CREATE this data connection' in str(ex_ctx.exception))
                else:
                    drc = TDataRegConnection()
                    drc.name = CXN
                    drc.type = "JDBC"
                    drc.jdbc_driver = "mysql"
                    drc.host = "cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com"
                    drc.port = 3306
                    drc.user_key = "awsps:///mysql/username"
                    drc.password_key = "awsps:///mysql/password"
                    drc.default_catalog = "jdbc_test"
                    drc.is_active = True
                    for user, can_create in [(USER, True), (USER_NO_ACCESS, False)]:
                        ctx.enable_token_auth(token_str=user)
                        if can_create:
                            create_connection(conn, drc)

                            # Since we did create as owner, we can drop it
                            delete_connection(conn, CXN)
                        else:
                            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                                create_connection(conn, drc)
                            self.assertTrue('CREATE this data connection' in str(ex_ctx.exception))

    def test_dataconnection_use_table(self):
        CXN = 'test_privileges_connection_use_table'
        ROLE = "test_privileges_connection_role"
        DB1 = "test_privileges_connection_db"
        TBL1 = "%s.tbl" % DB1
        USER = "testprivuser1"

        # Set up the NO_ACCESS user
        DB2 = "test_privileges_connection_db2"
        TBL2 = "%s.tbl2" % DB2
        ROLE2 = "test_privileges_connection_role2"
        USER_NO_ACCESS = "testprivuser2"

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s cascade" % (DB1),
                "DROP DATABASE IF EXISTS %s cascade" % (DB2),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (ROLE2),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "CREATE ROLE %s" % (ROLE2),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),
                "GRANT ROLE %s TO GROUP %s" % (ROLE2, USER_NO_ACCESS),

                # Create the DB
                "CREATE DATABASE %s" % (DB1),
                "CREATE DATABASE %s" % (DB2),

                # Create the connection and grant USE on it
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN),
                "GRANT USE ON DATACONNECTION %s TO ROLE %s" % (CXN, ROLE),

                # TODO: FIXME: why does this not work?
                #"GRANT CREATE_AS_OWNER ON DATABASE %s TO ROLE %s" % (DB1, ROLE),
                "GRANT CREATE ON CATALOG TO ROLE %s" % (ROLE),

                # Give the no access user only the ability to create but not use
                # TODO: FIXME: why does this not work?
                #"GRANT CREATE_AS_OWNER ON DATABASE %s TO ROLE %s" % (DB2, ROLE2),
                "GRANT CREATE ON CATALOG TO ROLE %s" % (ROLE2),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            for user, can_use, tbl in [(USER, True, TBL1), (USER_NO_ACCESS, False, TBL2)]:
                create_stmt = """CREATE EXTERNAL TABLE %s STORED as JDBC
                        TBLPROPERTIES(
                        'driver' = 'mysql',
                        'okera.connection.name' = '%s',
                        'jdbc.db.name'='jdbc_test',
                        'jdbc.schema.name'='public',
                        'table' = 'filter_pushdown_test'
                    )""" % (tbl, CXN)
                create_as_view_stmt = "%s USING VIEW AS 'select 1 as x'" % create_stmt

                ctx.enable_token_auth(token_str=user)
                if can_use:
                    conn.execute_ddl(create_stmt)

                    # Now we drop it and create a table using AS VIEW
                    conn.execute_ddl('DROP TABLE %s' % tbl)
                    conn.execute_ddl(create_as_view_stmt)
                else:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.execute_ddl(create_stmt)
                    self.assertTrue('USE the connection' in str(ex_ctx.exception))

                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.execute_ddl(create_as_view_stmt)
                    self.assertTrue('USE the connection' in str(ex_ctx.exception))

    def test_dataconnection_use_database(self):
        CXN = 'test_privileges_connection_use_database'
        ROLE = "test_privileges_connection_role"
        DB1 = "test_privileges_connection_db"
        USER = "testprivuser1"

        # Set up the NO_ACCESS user
        DB2 = "test_privileges_connection_db2"
        ROLE2 = "test_privileges_connection_role2"
        USER_NO_ACCESS = "testprivuser2"

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s cascade" % (DB1),
                "DROP DATABASE IF EXISTS %s cascade" % (DB2),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (ROLE2),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "CREATE ROLE %s" % (ROLE2),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),
                "GRANT ROLE %s TO GROUP %s" % (ROLE2, USER_NO_ACCESS),

                # Create the connection and grant USE on it
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN),
                "GRANT USE ON DATACONNECTION %s TO ROLE %s" % (CXN, ROLE),

                "GRANT CREATE_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE),

                # Give the no access user only the ability to create but not use
                "GRANT CREATE_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE2),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            for user, can_use, db in [(USER, True, DB1), (USER_NO_ACCESS, False, DB2)]:
                create_stmt = """CREATE DATABASE %s DBPROPERTIES
                                 ('okera.connection.name' = '%s');""" % (db, CXN)
                ctx.enable_token_auth(token_str=user)
                if can_use:
                    conn.execute_ddl(create_stmt)
                else:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.execute_ddl(create_stmt)
                    self.assertTrue('USE the connection' in str(ex_ctx.exception))

    def test_dataconnection_use_database_alter(self):
        CXN = 'test_privileges_connection_alter'
        ROLE = "test_privileges_connection_role"
        DB1 = "test_privileges_connection_db"
        USER = "testprivuser1"

        ROLE2 = "test_privileges_connection_role2"
        USER_NO_ACCESS = "testprivuser2"

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s cascade" % (DB1),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (ROLE2),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "CREATE ROLE %s" % (ROLE2),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),
                "GRANT ROLE %s TO GROUP %s" % (ROLE2, USER_NO_ACCESS),

                # Create the connection and grant USE on it
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN),
                """CREATE DATABASE %s DBPROPERTIES (
                    'okera.connection.name' = '%s');""" % (DB1, CXN),

                "GRANT USE ON DATACONNECTION %s TO ROLE %s" % (CXN, ROLE),
                "GRANT ALTER ON DATABASE %s TO ROLE %s" % (DB1, ROLE),

                # Give the no access user only the ability to ALTER but not use
                "GRANT ALTER ON DATABASE %s TO ROLE %s" % (DB1, ROLE2),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            for user, can_alter in [(USER, True), (USER_NO_ACCESS, False)]:
                alter_stmt = "ALTER DATABASE %s LOAD DEFINITIONS ()" % (DB1)
                alter_props_stmt = "ALTER DATABASE %s SET DBPROPERTIES ('okera.connection.name'='%s')" % (DB1, CXN)
                ctx.enable_token_auth(token_str=user)
                if can_alter:
                    conn.execute_ddl(alter_stmt)
                    conn.execute_ddl(alter_props_stmt)
                else:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.execute_ddl(alter_stmt)
                    self.assertTrue('USE the connection' in str(ex_ctx.exception))
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.execute_ddl(alter_props_stmt)
                    self.assertTrue('USE the connection' in str(ex_ctx.exception))

    def test_dataconnection_use_crawler(self):
        CXN = 'test_privileges_connection_crawler'
        ROLE = "test_privileges_connection_role"
        DB1 = "test_privileges_connection_db"
        USER = "testprivuser1"

        ROLE2 = "test_privileges_connection_role2"
        USER_NO_ACCESS = "testprivuser2"

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % (DB1),
                "DROP DATABASE IF EXISTS _okera_crawler_%s CASCADE" % (CXN),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (ROLE2),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "CREATE ROLE %s" % (ROLE2),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),
                "GRANT ROLE %s TO GROUP %s" % (ROLE2, USER_NO_ACCESS),

                # Create the connection and grant USE on it
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN),
                """CREATE DATABASE %s DBPROPERTIES (
                    'okera.connection.name' = '%s');""" % (DB1, CXN),

                "GRANT USE ON DATACONNECTION %s TO ROLE %s" % (CXN, ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            for user, can_use_crawler in [(USER, True), (USER, True)]:
                ctx.enable_token_auth(token_str=user)
                if can_use_crawler:
                    discover_crawler(conn, CXN)
                    manage_crawler(conn, CXN)
                else:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        discover_crawler(conn, CXN)
                        manage_crawler(conn, CXN)
                    self.assertTrue('USE this data connection' in str(ex_ctx.exception))

    def test_dataconnection_privileges_while_listing(self):
        CXN = 'test_privileges_connection_listing_cxn'
        CXN2 = 'test_privileges_connection_listing_cxn2'

        ROLE = "test_privileges_connection_listing_role"
        USER = "test_privileges_connection_listing_user1"

        ROLE2 = "test_privileges_connection_listing_role2"
        USER_NO_ACCESS = "test_privileges_connection_listing_user2"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:

            ddls = [
                "DROP DATACONNECTION IF EXISTS %s" % (CXN),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN2),
                "DROP ROLE IF EXISTS %s" % (ROLE),
                "DROP ROLE IF EXISTS %s" % (ROLE2),

                # Create the connection
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN),

                # Create the connection
                """CREATE DATACONNECTION %s CXNPROPERTIES
                    (
                    'connection_type'='JDBC',
                    'jdbc_driver'='mysql',
                    'host'='cerebro-db-test-long-running.cyn8yfvyuugz.us-west-2.rds.amazonaws.com',
                    'port'='3306',
                    'user_key'='awsps:///mysql/username',
                    'password_key'='awsps:///mysql/password',
                    'jdbc.db.name'='jdbc_test'
                    )
                """ % (CXN2),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE),
                "GRANT ROLE %s TO GROUP %s" % (ROLE, USER),

                "CREATE ROLE %s" % (ROLE2),
                "GRANT ROLE %s TO GROUP %s" % (ROLE2, USER_NO_ACCESS),

                "GRANT %s ON CATALOG TO ROLE %s" % ('CREATE_AS_OWNER', ROLE),
                "GRANT %s ON CATALOG TO ROLE %s" % ('CREATE_DATACONNECTION_AS_OWNER', ROLE),

                "GRANT %s ON DATACONNECTION %s TO ROLE %s" % ('SHOW', CXN, ROLE),
                "GRANT %s ON DATACONNECTION %s TO ROLE %s" % ('DROP', CXN, ROLE),
                "GRANT %s ON DATACONNECTION %s TO ROLE %s" % ('ALTER', CXN, ROLE),
                "GRANT %s ON DATACONNECTION %s TO ROLE %s" % ('USE', CXN, ROLE),

                "GRANT %s ON DATACONNECTION %s TO ROLE %s" % ('ALL', CXN2, ROLE),
                "GRANT %s ON DATACONNECTION %s TO ROLE %s" % ('USE', CXN2, ROLE)
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Verify that the root user (admin) can see both connections
            # If authorization is disabled or if user is an admin or if it's an internal
            # request, the user will always have 'ALL' permission
            ctx.disable_auth()
            connections = conn.manage_data_reg_connection("LIST", EMPTY_DRC, connection_pattern=CXN).connections
            assert len(connections) == 2
            sorted_names = sorted([connection.name for connection in connections])
            assert sorted_names[0] == CXN
            assert sorted_names[1] == CXN2
            for connection in connections:
                if connection.name in [CXN,CXN2]:
                    assert TAccessPermissionLevel.ALL in connection.access_levels

            ctx.enable_token_auth(token_str=USER)

            connections = conn.manage_data_reg_connection("LIST", EMPTY_DRC, connection_pattern=CXN).connections
            assert len(connections) == 2
            sorted_names = sorted([connection.name for connection in connections])
            assert sorted_names[0] == CXN
            assert sorted_names[1] == CXN2

            for connection in connections:
                if connection.name == CXN:
                    assert TAccessPermissionLevel.CREATE_DATACONNECTION_AS_OWNER not in connection.access_levels
                    assert TAccessPermissionLevel.SHOW in connection.access_levels
                    assert TAccessPermissionLevel.DROP in connection.access_levels
                    assert TAccessPermissionLevel.ALTER in connection.access_levels
                    assert TAccessPermissionLevel.USE in connection.access_levels
                if connection.name == CXN2:
                    assert TAccessPermissionLevel.CREATE_DATACONNECTION_AS_OWNER not in connection.access_levels
                    assert TAccessPermissionLevel.ALL in connection.access_levels
                    assert TAccessPermissionLevel.USE in connection.access_levels

            # Verify that a user with no access can't see the connection
            ctx.enable_token_auth(token_str=USER_NO_ACCESS)
            connections = conn.manage_data_reg_connection("LIST", EMPTY_DRC, connection_pattern=CXN).connections
            assert len(connections) == 0

    def test_dataconnection_connection_name_validation(self):
        CXN1 = 'test_connection_name_with_special_char_*'
        CXN2 = 'test_connection_name_with_special_char_$'
        CXN3 = 'test_connection_name_with_hyphen-'
        CXN4 = 'test_connection_name_   with_spaces_in_between'
        CXN5 = '    test_connection_name_with_leading_or_trailing_spaces    '
        VALID_CXN_NAME = 'test_valid_CONNECTION_name_123'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                drc = self._get_mysqltest_data_reg_connection_obj(CXN1)
                conn.manage_data_reg_connection('CREATE', drc)
            self.assertTrue(
                "Connection names can only contain characters, numbers and '_'"
                in str(ex_ctx.exception))

            with self.assertRaises(TRecordServiceException) as ex_ctx:
                drc = self._get_mysqltest_data_reg_connection_obj(CXN2)
                conn.manage_data_reg_connection('CREATE', drc)
            self.assertTrue(
                "Connection names can only contain characters, numbers and '_'"
                in str(ex_ctx.exception))

            with self.assertRaises(TRecordServiceException) as ex_ctx:
                drc = self._get_mysqltest_data_reg_connection_obj(CXN3)
                conn.manage_data_reg_connection('CREATE', drc)
            self.assertTrue(
                "Connection names can only contain characters, numbers and '_'"
                in str(ex_ctx.exception))

            with self.assertRaises(TRecordServiceException) as ex_ctx:
                drc = self._get_mysqltest_data_reg_connection_obj(CXN4)
                conn.manage_data_reg_connection('CREATE', drc)
            self.assertTrue(
                "Connection names can only contain characters, numbers and '_'"
                in str(ex_ctx.exception))

            with self.assertRaises(TRecordServiceException) as ex_ctx:
                drc = self._get_mysqltest_data_reg_connection_obj(CXN5)
                conn.manage_data_reg_connection('CREATE', drc)
            self.assertTrue(
                "Connection names can only contain characters, numbers and '_'"
                in str(ex_ctx.exception))

            drc = self._get_mysqltest_data_reg_connection_obj(VALID_CXN_NAME)
            drcs = conn.manage_data_reg_connection('CREATE', drc)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertEqual(drcs.connections[0].name, VALID_CXN_NAME,
                             "Test for valid connection name failed.")
            # Delete connection after the test
            drcs = conn.manage_data_reg_connection('DELETE', drc)
            self.assertTrue(len(drcs.connections) == 0)

    def test_dataconnection_delegated_test_connection(self):
        CXN1 = 'test_privileges_connection_test1'
        CXN2 = 'test_privileges_connection_test2'
        CXN3 = 'test_privileges_connection_test3'
        ROLE1 = 'test_privileges_connection_role1'
        ROLE2 = 'test_privileges_connection_role2'
        USER1 = "test_privileges_connection_user1"
        USER2 = "test_privileges_connection_user2"
        USER_NO_ACCESS = "test_privileges_connection_user_no_access"
        ROOT = 'root'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            ddls = [
                "DROP DATACONNECTION IF EXISTS %s" % (CXN1),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN2),
                "DROP DATACONNECTION IF EXISTS %s" % (CXN3),
                "DROP ROLE IF EXISTS %s" % (ROLE1),
                "DROP ROLE IF EXISTS %s" % (ROLE2),

                # Create the role and grant it to the user
                "CREATE ROLE %s" % (ROLE1),
                "GRANT ROLE %s TO GROUP %s" % (ROLE1, USER1),
                "CREATE ROLE %s" % (ROLE2),
                "GRANT ROLE %s TO GROUP %s" % (ROLE2, USER2),

                # Grant the user the right permissions
                "GRANT CREATE_DATACONNECTION_AS_OWNER ON CATALOG TO ROLE %s" % (ROLE1),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # USER1 should be able to test the connection while creating it as it has
            # CREATE_DATACONNECTION_AS_OWNER access on the CATALOG
            ctx.enable_token_auth(token_str=USER1)
            drc_cxn1 = self._get_mysqltest_data_reg_connection_obj(CXN1)
            # test the connection
            drcs = conn.manage_data_reg_connection("TEST_CREATE", drc_cxn1)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].test_status == "SUCCESS")
            # create the connection after successful testing
            conn.manage_data_reg_connection("CREATE", drc_cxn1)

            # USER2 should not be able to test CXN1.
            # USER2 must have 'ALL' or at least 'USE' access on CXN1
            ctx.enable_token_auth(token_str=USER2)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection("TEST_EXISTING", drc_cxn1)
            self.assertTrue('The user must have the appropriate permissions' in str(ex_ctx.exception))

            # Grant 'USE' on CXN1 to USER2
            ctx.enable_token_auth(token_str=ROOT)
            conn.execute_ddl('GRANT USE ON DATACONNECTION %s TO ROLE %s' % (CXN1, ROLE2))

            # Now, USER2 should be able to test CXN1
            ctx.enable_token_auth(token_str=USER2)
            drcs = conn.manage_data_reg_connection("TEST_EXISTING", drc_cxn1)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].test_status == "SUCCESS")

            # USER2 should NOT be able to test the connection (CXN2) while creating it as
            # it does not have CREATE_DATACONNECTION_AS_OWNER or ALL access on the CATALOG
            ctx.enable_token_auth(token_str=USER2)
            drc_cxn2 = self._get_mysqltest_data_reg_connection_obj(CXN2)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection("TEST_CREATE", drc_cxn2)
            self.assertTrue('The user must have the appropriate permissions' in str(ex_ctx.exception))

            # Grant 'CREATE_DATACONNECTION_AS_OWNER' to USER2 on CATALOG
            ctx.enable_token_auth(token_str=ROOT)
            conn.execute_ddl('GRANT CREATE_DATACONNECTION_AS_OWNER ON CATALOG TO ROLE %s' % (ROLE2))

            # Now, USER2 should be able to test CXN2 while creating
            ctx.enable_token_auth(token_str=USER2)
            drcs = conn.manage_data_reg_connection("TEST_CREATE", drc_cxn2)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].test_status == "SUCCESS")
            # create the connection after successful testing
            conn.manage_data_reg_connection("CREATE", drc_cxn2)

            # USER2 should also be able to test CXN2 while editing
            # test fails as the host is invalid
            ctx.enable_token_auth(token_str=USER2)
            drc_cxn2.host = 'invalid_mysql_host'
            drcs = conn.manage_data_reg_connection("TEST_EDIT", drc_cxn2)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].test_status == "FAILED")

            # USER_NO_ACCESS should not be able to test any connection (CXN1 or CXN2)
            ctx.enable_token_auth(token_str=USER_NO_ACCESS)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection("TEST_EXISTING", drc_cxn1)
            self.assertTrue('The user must have the appropriate permissions' in str(ex_ctx.exception))

            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection("TEST_EXISTING", drc_cxn2)
            self.assertTrue('The user must have the appropriate permissions' in str(ex_ctx.exception))

            # If USER2 tries to test a connection with same name as CXN1 (which was created by USER1)
            # System throws an error asking user to use a different connection name.
            ctx.enable_token_auth(token_str=USER2)
            drc_cxn1 = self._get_mysqltest_data_reg_connection_obj(CXN1)
            # test the connection
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection("TEST_CREATE", drc_cxn1)
            self.assertTrue('Please use a different conneciton name.' in str(ex_ctx.exception))

            # 1. Lets create CXN3 with USER1.
            # 2. USER2 should not be able to test CXN3 even though it has
            #   'CREATE_DATACONNECTION_AS_OWNER' on the CATALOG because CXN3 is not
            #    created by USER2.
            # 3. USER2 must have 'ALL' or at least 'USE' access on CXN3
            ctx.enable_token_auth(token_str=USER1)
            drc_cxn3 = self._get_mysqltest_data_reg_connection_obj(CXN3)
            # create the connection
            conn.manage_data_reg_connection("CREATE", drc_cxn3)
            # try to test CXN2 with USER2, it would not be able to.
            ctx.enable_token_auth(token_str=USER2)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection("TEST_EXISTING", drc_cxn3)
            self.assertTrue('The user must have the appropriate permissions' in str(ex_ctx.exception))

    def test_if_exists_clause_when_drop_connection(self):
        """ 'IF EXISTS' clause can not be specified through API calls.
            This test just verifies that the overall 'drop/delete connection'
            functionality works as expected.
        """
        MYSQL_CXN = 'test_if_exists_clause_mysql_conn'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()

            # DELETE a connection which does not exist, it should throw an error
            drc = TDataRegConnection()
            drc.name = MYSQL_CXN
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection('DELETE', drc)
            self.assertTrue('Connection does not exist: %s' % MYSQL_CXN in str(ex_ctx.exception))

            # create a connection
            drc_cxn = self._get_mysqltest_data_reg_connection_obj(MYSQL_CXN)
            conn.manage_data_reg_connection("CREATE", drc_cxn)

            # ensure that the connection is successfully created
            drcs = conn.manage_data_reg_connection("GET", drc_cxn)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].name == MYSQL_CXN)

            # DELETE the newly created connection
            drcs = conn.manage_data_reg_connection("DELETE", drc_cxn)
            self.assertTrue(len(drcs.connections) == 0)

            # get again and it returns empty
            drcs = conn.manage_data_reg_connection("GET", drc_cxn)
            self.assertTrue(len(drcs.connections) == 0)

    def test_if_not_exists_clause_when_create_connection(self):
        """ 'IF NOT EXISTS' clause can not be specified through API calls.
            This test just verifies that the overall 'create connection'
            functionality works as expected.
        """
        MYSQL_CXN = 'test_if_not_exists_clause_mysql_conn'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ctx.disable_auth()
            # Delete if exists
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS %s' % MYSQL_CXN)

            # create a connection
            drc_cxn = self._get_mysqltest_data_reg_connection_obj(MYSQL_CXN)
            conn.manage_data_reg_connection("CREATE", drc_cxn)

            # ensure that the connection is successfully created
            drcs = conn.manage_data_reg_connection("GET", drc_cxn)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].name == MYSQL_CXN)

            # Try to create the same connection, it should throw an error.
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.manage_data_reg_connection("CREATE", drc_cxn)
            self.assertTrue('Connection already exists: %s' % MYSQL_CXN in str(ex_ctx.exception))

            # DELETE the connection
            drcs = conn.manage_data_reg_connection("DELETE", drc_cxn)
            self.assertTrue(len(drcs.connections) == 0)

            # Re-create the connection and it should work
            drc_cxn = self._get_mysqltest_data_reg_connection_obj(MYSQL_CXN)
            conn.manage_data_reg_connection("CREATE", drc_cxn)

            # ensure that the connection is successfully created
            drcs = conn.manage_data_reg_connection("GET", drc_cxn)
            self.assertTrue(len(drcs.connections) == 1)
            self.assertTrue(drcs.connections[0].name == MYSQL_CXN)

if __name__ == "__main__":
    unittest.main()



# Copyright 2017 Okera Inc. All Rights Reserved.
#
# Tests that should run on any configuration. The server auth can be specified
# as an environment variables before running this test.
# pylint: disable=bad-continuation,bad-indentation,global-statement,unused-argument
# pylint: disable=no-self-use
import time
import unittest
import json
import numpy

from okera._thrift_api import TTypeId

from okera.tests import pycerebro_test_common as common
import cerebro_common as cerebro

retry_count = 0

class BasicTest(common.TestBase):
    def test_sparse_data(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            df = planner.scan_as_pandas("rs.sparsedata")
            self.assertEqual(96, len(df), msg=df)
            self.assertEqual(68, df['age'].count(), msg=df)
            self.assertEqual(10.0, df['age'].min(), msg=df)
            self.assertEqual(96.0, df['age'].max(), msg=df)
            self.assertEqual(b'sjc', df['defaultcity'].max(), msg=df)
            self.assertEqual(86, df['description'].count(), msg=df)

    def test_nulls(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            df = planner.scan_as_pandas("select string_col from rs.alltypes_null")
            self.assertEqual(1, len(df), msg=df)
            self.assertTrue(numpy.isnan(df['string_col'][0]), msg=df)

            df = planner.scan_as_pandas(
                "select length(string_col) as c from rs.alltypes_null")
            self.assertEqual(1, len(df), msg=df)
            self.assertTrue(numpy.isnan(df['c'][0]), msg=df)

    def test_timestamp_functions(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("""
                select date_add('2009-01-01', 10) as c from okera_sample.sample""")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('2009-01-11 00:00:00.000', str(json[0]['c']), msg=json)
            self.assertEqual('2009-01-11 00:00:00.000', str(json[1]['c']), msg=json)

    def test_duplicate_cols(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("""
                select record, record from okera_sample.sample""")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record']),
                             msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record_2']),
                             msg=json)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("""
                select record, record as record_2, record from okera_sample.sample""")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record']),
                             msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record_2']),
                             msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record_2_2']),
                             msg=json)

    def test_large_decimals(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("select num from rs.large_decimals2")
            self.assertTrue(len(json) == 6, msg=json)
            self.assertEqual('9012248907891233.020304050670',
                             str(json[0]['num']), msg=json)
            self.assertEqual('2343.999900000000', str(json[1]['num']), msg=json)
            self.assertEqual('900.000000000000', str(json[2]['num']), msg=json)
            self.assertEqual('32.440000000000', str(json[3]['num']), msg=json)
            self.assertEqual('54.230000000000', str(json[4]['num']), msg=json)
            self.assertEqual('4525.340000000000', str(json[5]['num']), msg=json)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            df = planner.scan_as_pandas("select num from rs.large_decimals2")
            self.assertTrue(len(df) == 6, msg=df)
            self.assertEqual('9012248907891233.020304050670',
                             str(df['num'][0]), msg=df)
            self.assertEqual('2343.999900000000', str(df['num'][1]), msg=df)
            self.assertEqual('900.000000000000', str(df['num'][2]), msg=df)
            self.assertEqual('32.440000000000', str(df['num'][3]), msg=df)
            self.assertEqual('54.230000000000', str(df['num'][4]), msg=df)
            self.assertEqual('4525.340000000000', str(df['num'][5]), msg=df)

    def test_date(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("select * from datedb.date_csv")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('Robert', str(json[0]['name']), msg=json)
            self.assertEqual(100, json[0]['id'], msg=json)
            self.assertEqual('1980-01-01', str(json[0]['dob']), msg=json)
            self.assertEqual('Michelle', str(json[1]['name']), msg=json)
            self.assertEqual(200, json[1]['id'], msg=json)
            self.assertEqual('1991-12-31', str(json[1]['dob']), msg=json)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            pd = planner.scan_as_pandas("select * from datedb.date_csv")
            self.assertTrue(len(pd) == 2, msg=pd)
            self.assertEqual(b'Robert', pd['name'][0], msg=pd)
            self.assertEqual(100, pd['id'][0], msg=pd)
            self.assertEqual('1980-01-01', str(pd['dob'][0]), msg=pd)
            self.assertEqual(b'Michelle', pd['name'][1], msg=pd)
            self.assertEqual(200, pd['id'][1], msg=pd)
            self.assertEqual('1991-12-31', str(pd['dob'][1]), msg=pd)

    def test_scan_as_json_max_records(self):
        sql = "select * from okera_sample.sample"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json(sql, max_records=1, max_client_process_count=1)
            self.assertTrue(len(json) == 1, msg='max_records not respected')
            json = planner.scan_as_json(sql, max_records=100, max_client_process_count=1)
            self.assertTrue(len(json) == 2, msg='max_records not respected')

    def test_scan_as_pandas_max_records(self):
        sql = "select * from okera_sample.sample"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            pd = planner.scan_as_pandas(sql, max_records=1, max_client_process_count=1)
            self.assertTrue(len(pd.index) == 1, msg='max_records not respected')
            pd = planner.scan_as_pandas(sql, max_records=100, max_client_process_count=1)
            self.assertTrue(len(pd.index) == 2, msg='max_records not respected')

    def test_scan_retry(self):
        global retry_count

        sql = "select * from okera_sample.sample"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            # First a sanity check
            pd = planner.scan_as_pandas(sql, max_records=1, max_client_process_count=1)
            self.assertTrue(len(pd.index) == 1, msg='test_scan_retry sanity check failed')

            # Patch scan_as_pandas to throw an IOError 2 times
            retry_count = 0
            def test_hook_retry(func_name, retries, attempt):
                if func_name != "plan":
                    return
                global retry_count
                retry_count = retry_count + 1
                if attempt < 2:
                    raise IOError('Fake Error')

            planner.test_hook_retry = test_hook_retry
            pd = planner.scan_as_pandas(sql, max_records=1, max_client_process_count=1)

            assert(retry_count == 3) # count = 2 failures + 1 success
            self.assertTrue(len(pd.index) == 1, msg='Failed to get data with retries')

    def test_worker_retry(self):
        global retry_count

        ctx = common.get_test_context()
        with common.get_worker(ctx) as worker:
            # First a sanity check
            v = worker.get_protocol_version()
            self.assertEqual('1.0', v)

            # Patch get_protocol_version to throw an IOError 2 times
            retry_count = 0
            def test_hook_retry(func_name, retries, attempt):
                if func_name != "get_protocol_version":
                    return
                global retry_count
                retry_count = retry_count + 1
                if attempt < 2:
                    raise IOError('Fake Error')

            worker.test_hook_retry = test_hook_retry
            v = worker.get_protocol_version()

            assert(retry_count == 3) # count = 2 failures + 1 success
            self.assertEqual('1.0', v)

    def test_overwrite_file(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            planner.execute_ddl("DROP TABLE IF EXISTS rs.dim")
            planner.execute_ddl("""CREATE EXTERNAL TABLE rs.dim
                (country_id INT, country_name STRING, country_code STRING)
                ROW FORMAT DELIMITED FIELDS TERMINATED BY ','
                LOCATION 's3://cerebro-datasets/starschema_demo/country_dim/'
                TBLPROPERTIES ('skip.header.line.count'='1')""")

            # Copy one version of the file into the target location
            cerebro.run_shell_cmd('aws s3 cp ' +\
                's3://cerebro-datasets/country_dim_src/country_DIM.csv ' +\
                's3://cerebro-datasets/starschema_demo/country_dim/country_DIM.csv')
            before = planner.scan_as_json('rs.dim')[0]
            self.assertEqual("France", before['country_name'], msg=str(before))

            # Copy another version. This file has the same length but a different
            # character. S3 maintains time in ms timestamp, so sleep a bit.
            time.sleep(1)

            cerebro.run_shell_cmd('aws s3 cp ' +\
                's3://cerebro-datasets/country_dim_src/country_DIM2.csv ' +\
                's3://cerebro-datasets/starschema_demo/country_dim/country_DIM.csv')
            i = 0
            while i < 10:
                after = planner.scan_as_json('rs.dim')[0]
                if 'france' in after['country_name']:
                    return
                self.assertEqual("France", after['country_name'], msg=str(after))
                time.sleep(.1)
                i = i + 1
            self.fail(msg="Did not updated result in time.")

    def test_scan_as_json_newline_delimiters(self):
        sql1 = '''select
         *
        from
        okera_sample.sample'''
        sql2 = '''select
        *
        from
        okera_sample.sample'''
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json(sql1, max_records=100, max_client_process_count=1)
            self.assertTrue(
                len(json) == 2,
                msg='could parse query with newline and space delimiters')
            json = planner.scan_as_json(sql2, max_records=100, max_client_process_count=1)
            self.assertTrue(
                len(json) == 2,
                msg='could parse query with newline delimiters')

    def test_scan_as_json_using_with_clause(self):
        sql1 = '''WITH male_customers AS
         (SELECT * FROM okera_sample.users WHERE gender = 'M')
         SELECT * FROM male_customers;'''
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json(sql1, max_records=100, max_client_process_count=1)
            self.assertTrue(
                len(json) == 100,
                msg='could parse query that starts with "with"')

    def test_scan_as_json_serialization(self):
        sql = "select * from rs.alltypes"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json.loads(json.dumps(planner.scan_as_json(sql)))

    def test_das_6218(self):
        DB = "das_6218"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            self._create_all_types(conn, DB)

            # Verify schema from catalog, plan and data are all timestamp
            schema = conn.plan('%s.alltypes' % DB).schema
            self.assertEqual(schema.cols[10].type.type_id, TTypeId.TIMESTAMP_NANOS)
            df = conn.scan_as_pandas('SELECT timestamp_col FROM %s.alltypes' % DB)
            self.assertEqual(str(df.dtypes[0]), 'datetime64[ns, UTC]')
            catalog_schema = conn.list_datasets(DB, name='alltypes')[0].schema
            self.assertEqual(
                catalog_schema.cols[10].type.type_id, TTypeId.TIMESTAMP_NANOS)
            print(df)

            # Create a view that is proper with the explicit cast.
            conn.execute_ddl('''
                CREATE VIEW %s.v1(ts STRING)
                AS
                SELECT cast(timestamp_col AS STRING) FROM %s.alltypes''' % (DB, DB))
            # Verify schema from catalog, plan and data are all strings
            catalog_schema = conn.list_datasets(DB, name='v1')[0].schema
            self.assertEqual(catalog_schema.cols[0].type.type_id, TTypeId.STRING)
            schema = conn.plan('%s.v1' % DB).schema
            self.assertEqual(schema.cols[0].type.type_id, TTypeId.STRING)
            df = conn.scan_as_pandas('%s.v1' % DB)
            self.assertEqual(str(df.dtypes[0]), 'object')
            print(df)

            # We want to carefully construct a view that has mismatched types with
            # the view definition. The view just selects a timestamp columns but we
            # will force the catalog type to be string, instead of timestamp.
            # This forces the planner to produce an implicit cast.
            conn.execute_ddl('''
                CREATE EXTERNAL VIEW %s.v2(ts STRING)
                SKIP_ANALYSIS USING VIEW DATA AS
                "SELECT timestamp_col FROM %s.alltypes"''' % (DB, DB))
            # Verify schema from catalog, plan and data are all strings
            catalog_schema = conn.list_datasets(DB, name='v2')[0].schema
            self.assertEqual(catalog_schema.cols[0].type.type_id, TTypeId.STRING)
            schema = conn.plan('%s.v2' % DB).schema
            self.assertEqual(schema.cols[0].type.type_id, TTypeId.STRING)
            df = conn.scan_as_pandas('%s.v2' % DB)
            self.assertEqual(str(df.dtypes[0]), 'object')
            print(df)

    def test_zd_1633(self):
        DB = "zd_1633"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            conn.execute_ddl('''
                CREATE EXTERNAL TABLE %s.t(
                  s STRING)
                STORED AS TEXTFILE
                LOCATION 's3://cerebrodata-test/zd-1627/'
                ''' % DB)
            res = conn.scan_as_json('%s.t' % DB)
            # Default quote, only 1 row
            self.assertEqual(1, len(res))

            # No quote, should be 2 rows now
            conn.execute_ddl("ALTER TABLE %s.t SET SERDEPROPERTIES('quoteChar'='')" % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(2, len(res))

            # Recreate using table properties
            conn.execute_ddl('DROP TABLE %s.t' % DB)
            conn.execute_ddl('''
                CREATE EXTERNAL TABLE %s.t(
                  s STRING)
                STORED AS TEXTFILE
                LOCATION 's3://cerebrodata-test/zd-1627/'
                TBLPROPERTIES('okera.text-table.default-quote-char'='')
                ''' % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(2, len(res))

            # Explicit serde properties overrides table properties
            conn.execute_ddl('''
                ALTER TABLE %s.t SET SERDEPROPERTIES('quoteChar'='"')''' % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(1, len(res))

            # Table with two cols
            conn.execute_ddl('DROP TABLE %s.t' % DB)
            conn.execute_ddl('''
                CREATE EXTERNAL TABLE %s.t(
                  c1 STRING, c2 STRING)
                STORED AS TEXTFILE
                LOCATION 's3://cerebrodata-test/customers/c1/zd1633_2/'
                TBLPROPERTIES('skip.header.line.count'='1')
                ''' % DB)
            conn.execute_ddl(
                "ALTER TABLE %s.t SET SERDEPROPERTIES('field.delim'=',')" % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(1, len(res))

            # Remove quote handling
            conn.execute_ddl("ALTER TABLE %s.t SET SERDEPROPERTIES('quoteChar'='')" % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(2, len(res))
            self.assertEqual('"123', res[1]['c2'])

    def test_scan_delta_dummy_metadata(self):
        DB = "delta_dummy_scan_db"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            conn.execute_ddl('''
      CREATE EXTERNAL TABLE %s.airline_events_delta (
          col array<string>)
      WITH SERDEPROPERTIES (
          'path'='s3://cerebrodata-test/delta/airlines/airline_events/',
          'serialization.format'='1')
      STORED AS TEXTFILE
      LOCATION 's3://cerebrodata-test/delta/airlines/airline_events/'
      TBLPROPERTIES (
          'totalSize'='66153', 'numRows'='-1', 'rawDataSize'='-1',
          'okera.delta.infer-schema'='true',
          'COLUMN_STATS_ACCURATE'='false',
          'spark.sql.sources.schema.part.0'='{\"type\":\"struct\",\"fields\":[]}',
          'numFiles'='1', 'spark.sql.partitionProvider'='catalog',
          'spark.sql.sources.schema.numParts'='1', 'spark.sql.sources.provider'='delta',
          'spark.sql.create.version'='3.0.1')
                ''' % DB)
            res = conn.scan_as_json('%s.airline_events_delta' % DB)
            self.assertEqual(1326, len(res))

    def test_pandas_index(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            # We scan a small table but with a tiny min task size, so that it is
            # guaranteed to generate more than one task. We then verify that our
            # overall row indices are correct.
            df = planner.scan_as_pandas("okera_sample.sample", min_task_size=1)

            indices = []
            for i, row in df.iterrows():
                indices.append(i)

            assert len(indices) == 2
            assert indices[0] == 0
            assert indices[1] == 1

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for managing attributes
#
# pylint: disable=bad-continuation
# pylint: disable=too-many-locals

import unittest

from okera.tests import pycerebro_test_common as common

class AttributesTest(common.TestBase):
    @staticmethod
    def _contains_attribute(namespace, key, attributes):
        for attr in attributes:
            if attr.attribute_namespace == namespace and attr.key == key:
                return True
        return False

    def _verify_attr(self, val, namespace, key, db, tbl, col):
        self.assertEqual(val.attribute.attribute_namespace, namespace)
        self.assertEqual(val.attribute.key, key)
        self.assertEqual(val.database, db)
        self.assertEqual(val.table, tbl)
        self.assertEqual(val.column, col)

    def test_basic(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.create_attribute('abac_test', 'v1')
            conn.create_attribute('abac_test', 'v2')
            attributes = conn.list_attributes('abac_test')
            self.assertTrue(len(attributes) >= 2, msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v1', attributes),
                            msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v2', attributes),
                            msg=str(attributes))
            old_len = len(attributes)
            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test' in namespaces)

            # Delete the attribute and make sure it is gone
            self.assertTrue(conn.delete_attribute('abac_test', 'v2'))
            attributes = conn.list_attributes('abac_test')
            for a in attributes:
                self.assertTrue(a.id is not None)
            self.assertEqual(old_len, len(attributes) + 1)
            self.assertTrue(self._contains_attribute('abac_test', 'v1', attributes),
                            msg=str(attributes))
            self.assertFalse(self._contains_attribute('abac_test', 'v2', attributes),
                             msg=str(attributes))
            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test' in namespaces)

    def test_multiple_namespaces(self):
        test_user = "ATTRIBUTE_TEST_USER"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.create_attribute('abac_test1', 'v')
            conn.create_attribute('abac_test2', 'v')
            conn.create_attribute('abac_test3', 'v')
            conn.execute_ddl("DROP ROLE IF EXISTS attribute_test_role")

            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test1' in namespaces)
            self.assertTrue('abac_test2' in namespaces)
            self.assertTrue('abac_test3' in namespaces)

            # Editable only is the same for admin
            for editable in [True, False]:
                namespaces = conn.list_attribute_namespaces(editable_only=editable)
                self.assertTrue('abac_test1' in namespaces)
                self.assertTrue('abac_test2' in namespaces)
                self.assertTrue('abac_test3' in namespaces)

                attributes = conn.list_attributes('abac_test1')
                self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
                self.assertFalse(self._contains_attribute('abac_test2', 'v', attributes))

                # List in all namespaces
                attributes = conn.list_attributes(editable_only=editable)
                self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
                self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
                self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))

                attributes = conn.list_attributes(None, editable_only=editable)
                self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
                self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
                self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))

                attributes = conn.list_attributes('', editable_only=editable)
                self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
                self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
                self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))

            # Try as a test user, no edit access
            ctx.enable_token_auth(token_str=test_user)
            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test1' in namespaces)
            self.assertTrue('abac_test2' in namespaces)
            self.assertTrue('abac_test3' in namespaces)
            namespaces = conn.list_attribute_namespaces(editable_only=True)
            self.assertEqual(len(namespaces), 0)

            attributes = conn.list_attributes('abac_test1')
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertFalse(self._contains_attribute('abac_test2', 'v', attributes))
            attributes = conn.list_attributes('abac_test1', editable_only=True)
            self.assertEqual(len(attributes), 0)

            attributes = conn.list_attributes(editable_only=False)
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))
            attributes = conn.list_attributes('', editable_only=True)
            self.assertEqual(len(attributes), 0)

            # Grant this test user all on two namespaces
            ctx.disable_auth()
            conn.execute_ddl("CREATE ROLE attribute_test_role")
            conn.execute_ddl("GRANT ROLE attribute_test_role TO GROUP " + test_user)
            conn.execute_ddl(
                "GRANT ALL ON ATTRIBUTE NAMESPACE abac_test1 TO ROLE attribute_test_role")

            ctx.enable_token_auth(token_str=test_user)
            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test1' in namespaces)
            self.assertTrue('abac_test2' in namespaces)
            self.assertTrue('abac_test3' in namespaces)
            namespaces = conn.list_attribute_namespaces(editable_only=True)
            self.assertTrue('abac_test1' in namespaces)
            self.assertTrue('abac_test2' not in namespaces)

            attributes = conn.list_attributes('abac_test1')
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertFalse(self._contains_attribute('abac_test2', 'v', attributes))
            attributes = conn.list_attributes('abac_test1', editable_only=True)
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertFalse(self._contains_attribute('abac_test2', 'v', attributes))

            attributes = conn.list_attributes(editable_only=False)
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))
            attributes = conn.list_attributes('', editable_only=True)
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertFalse(self._contains_attribute('abac_test2', 'v', attributes))

    def test_create_delete(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.delete_attribute('abac_test', 'v1')
            conn.delete_attribute('abac_test', 'v2')

            # Try deleting again, should not exist
            self.assertFalse(conn.delete_attribute('abac_test', 'v1'))
            self.assertFalse(conn.delete_attribute('abac_test', 'v2'))

            conn.create_attribute('abac_test', 'v1')
            conn.create_attribute('abac_test', 'v2')
            attributes = conn.list_attributes('abac_test')
            self.assertTrue(len(attributes) >= 2, msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v1', attributes),
                            msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v2', attributes),
                            msg=str(attributes))

            # Try deleting should exist
            self.assertTrue(conn.delete_attribute('abac_test', 'v1'))
            self.assertTrue(conn.delete_attribute('ABAC_test', 'V2'))

    def test_assign_get_attributes(self):
        db = 'attributes_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.create_attribute('abac_test', 'V1')
            conn.create_attribute('ABAC_test', 'v2')

            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db)
            conn.execute_ddl('CREATE DATABASE %s' % db)
            self.assertEqual(len(conn.list_datasets(db)), 0)

            # Create 2 tables and 2 views
            conn.execute_ddl('CREATE TABLE %s.t1(c1 int, c2 int, c3 int)' % db)
            conn.execute_ddl('CREATE TABLE %s.t2(c1 int, c2 int, c3 int)' % db)
            conn.execute_ddl('CREATE VIEW %s.v1 AS SELECT * from %s.t1' % (db, db))
            conn.execute_ddl('CREATE VIEW %s.v2 AS SELECT * from %s.t1' % (db, db))
            self.assertEqual(len(conn.list_datasets(db)), 4)

            # Get the attributes on t1, should be empty
            ds = conn.list_datasets(db, name='t1')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs_by_col = self.collect_column_attributes(ds)
            self.assertTrue(not attrs_by_col)

            # Ensure get_tags returns empty
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.t1') as v" % db)[0]['v'])
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.t1.c1') as v" % db)[0]['v'])
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.t1.not_a_col') as v" % db)[0]['v'])

            # Assign abac_test.v1 to t1
            conn.assign_attribute('abac_TEST', 'v1', db, 't1', cascade=False)
            conn.assign_attribute('abac_test', 'v1', db, 't1', cascade=False)
            ds = conn.list_datasets(db, name='t1')[0]
            attrs = ds.attribute_values
            self.assertTrue(attrs is not None)
            self.assertEqual(len(attrs), 1)
            self._verify_attr(attrs[0], 'abac_test', 'v1', db, 't1', None)
            self.assertTrue(not self.collect_column_attributes(ds))

            # Check get_tags
            self.assertEqual('abac_test.v1',
                conn.scan_as_json("select get_tags('%s.t1') as v" % db)[0]['v'])
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.t1.c1') as v" % db)[0]['v'])

            # Assign abac_test.v2 to t1
            conn.assign_attribute('abac_test', 'v2', db, 't1', cascade=False)
            ds = conn.list_datasets(db, name='t1')[0]
            attrs = ds.attribute_values
            self.assertTrue(attrs is not None)
            self.assertEqual(len(attrs), 2)
            if attrs[0].attribute.key == 'v1':
                self._verify_attr(attrs[0], 'abac_test', 'v1', db, 't1', None)
                self._verify_attr(attrs[1], 'abac_test', 'v2', db, 't1', None)
            else:
                self._verify_attr(attrs[1], 'abac_test', 'v1', db, 't1', None)
                self._verify_attr(attrs[0], 'abac_test', 'v2', db, 't1', None)

            # Check get_tags
            self.assertEqual('abac_test.v1,abac_test.v2',
                conn.scan_as_json("select get_tags('%s.t1') as v" % db)[0]['v'])

            # Assign abac_test.v1 to v1 and v2.c2
            conn.assign_attribute('abac_test', 'v1', db, 'v1')
            conn.assign_attribute('abac_test', 'v1', db, 'v2', 'c2')
            ds = conn.list_datasets(db, name='v1')[0]
            self._verify_attr(ds.attribute_values[0], 'abac_test', 'v1', db, 'v1', None)

            # Check get_tags
            self.assertEqual('abac_test.v1',
                conn.scan_as_json("select get_tags('%s.v1') as v" % db)[0]['v'])
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.v2.c1') as v" % db)[0]['v'])
            self.assertEqual('abac_test.v1',
                conn.scan_as_json("select get_tags('%s.v2.c2') as v" % db)[0]['v'])

            # Test get_tags on the table. The tag is on a column in the table but
            # we want this to count.
            self.assertEqual('abac_test.v1',
                conn.scan_as_json("select get_tags('%s.v2') as v" % db)[0]['v'])
            self.assertTrue(conn.scan_as_json(
                "select has_tag('%s.v2', 'abac_test.v1') as v" % db)[0]['v'])

            # Check has_tag
            self.assertTrue(conn.scan_as_json(
                "select has_tag('%s.t1', 'abac_test.v1') as v" % db)[0]['v'])
            self.assertTrue(conn.scan_as_json(
                "select has_tag('%s.t1', 'abac_test.v2') as v" % db)[0]['v'])
            self.assertFalse(conn.scan_as_json(
                "select has_tag('%s.t1', 'abac_test.v3') as v" % db)[0]['v'])
            self.assertTrue(conn.scan_as_json(
                "select has_tag('%s.v1', 'abac_test.v1') as v" % db)[0]['v'])
            self.assertFalse(conn.scan_as_json(
                "select has_tag('%s.v2.c1', 'abac_test.v1') as v" % db)[0]['v'])
            self.assertTrue(conn.scan_as_json(
                "select has_tag('%s.v2.c2', 'abac_test.v1') as v" % db)[0]['v'])
            self.assertFalse(conn.scan_as_json(
                "select has_tag('%s.v2.c1', 'abac_test.*') as v" % db)[0]['v'])
            self.assertTrue(conn.scan_as_json(
                "select has_tag('%s.v2.c2', 'abac_test.*') as v" % db)[0]['v'])

            ds = conn.list_datasets(db, name='v2')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self.collect_column_attributes(ds)
            self.assertTrue(('%s.v2.c2' % db) in attrs)
            self.assertTrue(('%s.v2.c1' % db) not in attrs)

            # Try again with upper case db
            ds = conn.list_datasets(db.upper(), name='V2')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self.collect_column_attributes(ds)
            self.assertTrue(('%s.v2.c2' % db) in attrs)
            self.assertTrue(('%s.v2.c1' % db) not in attrs)

            # Unassign abac_test.v1 from v2.c2
            conn.unassign_attribute('ABAC_test', 'v1', db, 'v2', 'c2')
            ds = conn.list_datasets(db, name='v2')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self.collect_column_attributes(ds)
            self.assertTrue(not attrs)
            # Verify the assignment to v1 is still there
            ds = conn.list_datasets(db, name='v1')[0]
            self._verify_attr(ds.attribute_values[0], 'abac_test', 'v1', db, 'v1', None)

            # Check get_tags
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.v2') as v" % db)[0]['v'])
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.v2.c1') as v" % db)[0]['v'])
            self.assertEqual('',
                conn.scan_as_json("select get_tags('%s.v2.c2') as v" % db)[0]['v'])
            self.assertEqual('abac_test.v1,abac_test.v2',
                conn.scan_as_json("select get_tags('%s.t1') as v" % db)[0]['v'])

            # Unassign abac_test.v1 from v1
            conn.unassign_attribute('abac_test', 'V1', db, 'v1')
            ds = conn.list_datasets(db, name='v1')[0]
            self.assertTrue(ds.attribute_values is None)

    def test_assign_invalid_attributes(self):
        db = 'attributes_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.create_attribute('abac_test', 'V1')

            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db)
            conn.execute_ddl('CREATE DATABASE %s' % db)

            conn.assign_attribute('abac_test', 'v1', db)
            with self.assertRaises(Exception) as ex_ctx:
                conn.assign_attribute('abac_test', 'not-there', db)
            self.assertTrue('Cannot assign attributes' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

    @unittest.skip("This requires autotagging to be configured.")
    def test_tag_table(self):
        db = 'attributes_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db)
            conn.execute_ddl('CREATE DATABASE %s' % db)
            conn.execute_ddl(('CREATE EXTERNAL TABLE %s.test LIKE AVRO ' +\
                '"s3://cerebrodata-test/poc_chase/avrodata/" STORED AS AVRO ' +\
                'LOCATION "s3://cerebrodata-test/poc_chase/avrodata/"') % db)

            # Should have no tags
            ds = conn.list_datasets(db, name='test')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self.collect_column_attributes(ds)
            self.assertEqual(len(attrs), 0)

            # Execute tag command
            conn.execute_ddl('ALTER TABLE %s.test EXECUTE AUTOTAG' % db)

            # Should have tags
            ds = conn.list_datasets(db, name='test')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self.collect_column_attributes(ds)
            self.assertTrue('%s.test.email' % db in attrs)

    def test_cascade(self):
        db = 'attributes_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.create_attribute('abac_test', 'v')
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db)
            conn.execute_ddl('CREATE DATABASE %s' % db)
            conn.execute_ddl('CREATE TABLE %s.tbl(i int)' % db)
            conn.execute_ddl('CREATE TABLE %s.tbl2(i int)' % db)
            conn.execute_ddl('CREATE VIEW %s.v1 AS SELECT * FROM %s.tbl' % (db, db))
            conn.execute_ddl('CREATE VIEW %s.v2 AS SELECT * FROM %s.tbl' % (db, db))
            conn.execute_ddl('CREATE VIEW %s.v1_1 AS SELECT * FROM %s.v1' % (db, db))
            tag = 'abac_test.v'

            for ddl, explicit in \
                    [(True, True), (True, False), (False, True), (False, False)]:
                # Should have no tags
                for v in ['tbl', 'tbl2', 'v1', 'v2', 'v1_1']:
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is None)

                # Assign tag and cascade. Everything based on tbl should have tags
                if ddl:
                    if explicit:
                        conn.execute_ddl('ALTER TABLE %s.tbl ADD ATTRIBUTE %s CASCADE' % \
                            (db, tag))
                    else:
                        conn.execute_ddl('ALTER TABLE %s.tbl ADD ATTRIBUTE %s ' % \
                            (db, tag))
                else:
                    if explicit:
                        conn.assign_attribute('abac_test', 'v', db, 'tbl', cascade=True)
                    else:
                        conn.assign_attribute('abac_test', 'v', db, 'tbl')

                for v in ['tbl', 'v1', 'v2', 'v1_1']:
                    print("Running it on " + v)
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is not None)
                for v in ['tbl2']:
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is None)

                ## Unassign from v1, should cascade to v1_1
                conn.execute_ddl('ALTER VIEW %s.v1 DROP ATTRIBUTE %s CASCADE' % (db, tag))
                ## These should have tags
                for v in ['tbl', 'v2']:
                    print("Running it on " + v)
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is not None)
                ## These should not have tags
                for v in ['v1', 'v1_1']:
                    print("Running it on " + v)
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is None)

                # Unassign from root, nothing should have tags now
                if ddl:
                    if explicit:
                        conn.execute_ddl('ALTER TABLE %s.tbl DROP ATTRIBUTE %s CASCADE' %\
                            (db, tag))
                    else:
                        conn.execute_ddl('ALTER TABLE %s.tbl DROP ATTRIBUTE %s ' % \
                            (db, tag))
                else:
                    if explicit:
                        conn.unassign_attribute('abac_test', 'v', db, 'tbl', cascade=True)
                    else:
                        conn.unassign_attribute('abac_test', 'v', db, 'tbl')

                for v in ['tbl', 'tbl2', 'v1', 'v2', 'v1_1']:
                    print("Running it on " + v)
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is None)

                # Assign without cascading, should only be on table
                if ddl:
                    conn.execute_ddl(
                        'ALTER TABLE %s.tbl ADD ATTRIBUTE %s DO NOT CASCADE' %
                        (db, tag))
                else:
                    conn.assign_attribute('abac_test', 'v', db, 'tbl', cascade=False)

                for v in ['tbl']:
                    print("Running it on " + v)
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is not None)
                for v in ['tbl2', 'v1', 'v2', 'v1_1']:
                    ds = conn.list_datasets(db, name=v)[0]
                    self.assertTrue(ds.attribute_values is None)

                if ddl:
                    conn.execute_ddl(
                        'ALTER TABLE %s.tbl DROP ATTRIBUTE %s DO NOT CASCADE' %
                        (db, tag))
                else:
                    conn.unassign_attribute('abac_test', 'v', db, 'tbl', cascade=False)

    def _test_cascade_case(self, conn, db, view_def, c1, c2, no_c1=False, no_c2=False):
        conn.execute_ddl('DROP VIEW IF EXISTS %s.v' % db)

        # Create the view and verify inheritance
        conn.execute_ddl(view_def)
        ds = conn.list_datasets(db, name='v')[0]
        attrs = self.collect_column_attributes(ds)
        if no_c1:
            self.assertTrue('attr_test_db.v.' + c1 not in attrs)
        else:
            self.assertEqual(attrs['attr_test_db.v.' + c1], ['attr_test_db.a1'])
        if no_c2:
            self.assertTrue('attr_test_db.v.' + c2 not in attrs)
        else:
            self.assertEqual(attrs['attr_test_db.v.' + c2], ['attr_test_db.a2'])

        # Assign a tag to verify cascade
        conn.execute_ddl("alter table %s.base_table add column attribute c1 %s.a3"\
              % (db, db))
        ds = conn.list_datasets(db, name='v')[0]
        attrs = self.collect_column_attributes(ds)
        if no_c1:
            self.assertTrue('attr_test_db.v.' + c1 not in attrs)
        else:
            # C1 should have both tags, a3 from assignment after view creation
            self.assertEqual(attrs['attr_test_db.v.' + c1],
                             ['attr_test_db.a1', 'attr_test_db.a3'])
        if no_c2:
            self.assertTrue('attr_test_db.v.' + c2 not in attrs)
        else:
            self.assertEqual(attrs['attr_test_db.v.' + c2], ['attr_test_db.a2'])

        conn.execute_ddl("alter table %s.base_table drop column attribute c1 %s.a3"\
              % (db, db))

    def test_cascade_cases(self):
        db = 'attr_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl("create attribute if not exists %s.a1" % db)
            conn.execute_ddl("create attribute if not exists %s.a2" % db)
            conn.execute_ddl("create attribute if not exists %s.a3" % db)

            conn.execute_ddl("drop database if exists %s cascade" % db)
            conn.execute_ddl("create database if not exists %s" % db)

            # Create base table and intermediate view
            conn.execute_ddl(("create table %s.base_table " +
                "(c1 int attribute %s.a1, c2 int attribute %s.a2)") % (db, db, db))
            conn.execute_ddl(
                "create view %s.base_view as select * from %s.base_table" % (db, db))

            ds = conn.list_datasets(db, name='base_table')[0]
            attrs = self.collect_column_attributes(ds)
            self.assertEqual(attrs['attr_test_db.base_table.c1'], ['attr_test_db.a1'])
            self.assertEqual(attrs['attr_test_db.base_table.c2'], ['attr_test_db.a2'])

            self._test_cascade_case(conn, db,
                "create view %s.v as select * from %s.base_table" % (db, db),
                'c1', 'c2')
            self._test_cascade_case(conn, db,
                "create view %s.v as select c1 as c3, c2 as c4 from %s.base_table" % \
                    (db, db),
                'c3', 'c4')
            self._test_cascade_case(conn, db,
                "create view %s.v as select c1 as c2, c2 as c1 from %s.base_table" % \
                    (db, db),
                'c2', 'c1')
            # Filters should have no impact
            self._test_cascade_case(conn, db,
                ("create view %s.v as select c1 as c2, c2 as c1 from %s.base_table " +
                "where c1 is not null") % (db, db),
                'c2', 'c1')

            # Inheritance through joins
            self._test_cascade_case(conn, db,
                ("create view %s.v as select a.c1, b.c2 " +\
                "from %s.base_view a join %s.base_view b") % (db, db, db),
                'c1', 'c2')

            # Expressions stop cascade
            self._test_cascade_case(conn, db,
                "create view %s.v as select c1 + 1 as c1, c2 from %s.base_table" % \
                    (db, db),
                'c1', 'c2', True)

            # Inheritance through aggs
            # FIXME: not working
            #self._test_cascade_case(conn, db,
            #    ("create view %s.v as select count(c1) as c1, c2 " +\
            #    "from %s.base_table group by c2") % (db, db),
            #    'c1', 'c2', True)
            #self._test_cascade_case(conn, db,
            #    ("create view %s.v as select count(c1) as c1, c2 " +\
            #    "from %s.base_table group by 2") % (db, db),
            #    'c1', 'c2', True)

            # Inheritance through unions
            # FIXME: not working
            #self._test_cascade_case(conn, db,
            #    ("create view %s.v as select c1, c2 FROM %s.base_table " +\
            #    "UNION ALL SELECT c1, c2 FROM %s.base_table") % (db, db, db),
            #    'c1', 'c2')
            #self._test_cascade_case(conn, db,
            #    ("create view %s.v as select c1, c2 FROM %s.base_table " +\
            #    "UNION ALL SELECT c2, c1 FROM %s.base_table") % (db, db, db),
            #    'c1', 'c2')

            # Add sort
            # FIXME: not working
            #self._test_cascade_case(conn, db,
            #    "create view %s.v as select * from %s.base_table order by c1" % (db, db),
            #    'c1', 'c2')

    def test_nested_struct_attributes(self):
        db = 'attr_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl("create attribute if not exists %s.a1" % db)
            conn.execute_ddl("create attribute if not exists %s.a2" % db)
            conn.execute_ddl("create attribute if not exists %s.a3" % db)

            expected_tbl_leaf_describe = '''
+------+----------------+---------+-----------------+
| name |      type      | comment |    attributes   |
+------+----------------+---------+-----------------+
|  id  |     bigint     |         | attr_test_db.a2 |
|  s1  |    struct<     |         |                 |
|      |    f1:string,  |         | attr_test_db.a1 |
|      |    s2:struct<  |         |                 |
|      |     f2:bigint, |         | attr_test_db.a3 |
|      |     f3:string  |         | attr_test_db.a1 |
|      |        >       |         |                 |
|      |       >        |         |                 |
+------+----------------+---------+-----------------+'''.strip()

            expected_tbl_describe = '''
+------+----------------+---------+-----------------+
| name |      type      | comment |    attributes   |
+------+----------------+---------+-----------------+
|  id  |     bigint     |         | attr_test_db.a2 |
|  s1  |    struct<     |         | attr_test_db.a3 |
|      |    f1:string,  |         | attr_test_db.a1 |
|      |    s2:struct<  |         | attr_test_db.a2 |
|      |     f2:bigint, |         | attr_test_db.a3 |
|      |     f3:string  |         | attr_test_db.a1 |
|      |        >       |         |                 |
|      |       >        |         |                 |
+------+----------------+---------+-----------------+'''.strip()

            expected_v1_describe = '''
+------+--------+---------+-----------------+
| name |  type  | comment |    attributes   |
+------+--------+---------+-----------------+
|  id  | bigint |         | attr_test_db.a2 |
|  f1  | string |         | attr_test_db.a1 |
|  f2  | bigint |         | attr_test_db.a3 |
|  f3  | string |         | attr_test_db.a1 |
+------+--------+---------+-----------------+
            '''.strip()

            expected_v2_describe = '''
+------+----------------+---------+-----------------+
| name |      type      | comment |    attributes   |
+------+----------------+---------+-----------------+
|  s1  |    struct<     |         | attr_test_db.a3 |
|      |    f1:string,  |         | attr_test_db.a1 |
|      |    s2:struct<  |         | attr_test_db.a2 |
|      |     f2:bigint, |         | attr_test_db.a3 |
|      |     f3:string  |         | attr_test_db.a1 |
|      |        >       |         |                 |
|      |       >        |         |                 |
|  s2  |    struct<     |         | attr_test_db.a2 |
|      |    f2:bigint,  |         | attr_test_db.a3 |
|      |    f3:string   |         | attr_test_db.a1 |
|      |       >        |         |                 |
+------+----------------+---------+-----------------+
            '''.strip()

            expected_v3_describe = '''
+------+--------------+---------+-----------------+
| name |     type     | comment |    attributes   |
+------+--------------+---------+-----------------+
|  id  |    bigint    |         | attr_test_db.a2 |
|  f   |    string    |         | attr_test_db.a1 |
|  s2  |   struct<    |         | attr_test_db.a2 |
|      |   f2:bigint, |         | attr_test_db.a3 |
|      |   f3:string  |         | attr_test_db.a1 |
|      |      >       |         |                 |
+------+--------------+---------+-----------------+
            '''.strip()

            expected_v4_describe = '''
+------+--------------+---------+-----------------+
| name |     type     | comment |    attributes   |
+------+--------------+---------+-----------------+
|  id  |    bigint    |         | attr_test_db.a2 |
|  s2  |   struct<    |         | attr_test_db.a2 |
|      |   f2:bigint, |         | attr_test_db.a3 |
|      |   f3:string  |         | attr_test_db.a1 |
|      |      >       |         |                 |
|  f2  |    bigint    |         | attr_test_db.a3 |
+------+--------------+---------+-----------------+
            '''.strip()

            expected_tbl_describe_formatted = '''
Table Attributes:    attr_test_db.a1
Columns With Attributes:
  id  attr_test_db.a2
  s1  attr_test_db.a3
  s1.f1  attr_test_db.a1
  s1.s2  attr_test_db.a2
  s1.s2.f2  attr_test_db.a3
  s1.s2.f3  attr_test_db.a1'''.strip()

            for cascade in [True, False]:
                conn.execute_ddl("drop database if exists %s cascade" % db)
                conn.execute_ddl("create database if not exists %s" % db)

                conn.execute_ddl('''
                    CREATE TABLE %s.tbl(
                      id BIGINT,
                      s1 STRUCT <f1: STRING, s2: STRUCT<f2: BIGINT, f3: STRING> >
                    ) stored as avro''' % db)

                if cascade:
                    # If cascade, create the view first.
                    conn.execute_ddl('''
                        CREATE VIEW %s.v1 AS SELECT
                          id, s1.f1 as f1, s1.s2.f2 as f2, s1.s2.f3 as f3
                        FROM %s.tbl''' % (db, db))
                    conn.execute_ddl('''
                        CREATE VIEW %s.v2 AS SELECT
                          s1, s1.s2 as s2
                        FROM %s.tbl''' % (db, db))
                    conn.execute_ddl('''
                        CREATE VIEW %s.v3 AS SELECT
                          id, s1.f1 as f, s1.s2 as s2
                        FROM %s.tbl''' % (db, db))
                    conn.execute_ddl('''
                        CREATE VIEW %s.v4 AS SELECT
                          id, s1.s2 as s2, s1.s2.f2 as f2
                        FROM %s.tbl''' % (db, db))

                # Assign tags to all the columns
                conn.assign_attribute(db, 'a1', db, 'tbl', cascade=cascade)
                conn.assign_attribute(db, 'a2', db, 'tbl', 'id', cascade=cascade)
                conn.assign_attribute(db, 'a1', db, 'tbl', 's1.f1', cascade=cascade)
                conn.assign_attribute(db, 'a3', db, 'tbl', 's1.s2.f2', cascade=cascade)
                conn.assign_attribute(db, 'a1', db, 'tbl', 's1.s2.f3', cascade=cascade)
                # Check the tags at just the leaf levels
                print(str(conn.execute_ddl_table_output('DESCRIBE %s.tbl' % db)))
                self.assertEqual(
                    expected_tbl_leaf_describe,
                    str(conn.execute_ddl_table_output('DESCRIBE %s.tbl' % db)))

                conn.assign_attribute(db, 'a3', db, 'tbl', 's1', cascade=cascade)
                conn.assign_attribute(db, 'a2', db, 'tbl', 's1.s2', cascade=cascade)

                ds = conn.list_datasets(db, name='tbl')[0]
                self.assertEqual('a1', ds.attribute_values[0].attribute.key)
                attrs = self.collect_column_attributes(ds)
                self.assertEqual(['attr_test_db.a2'], attrs['attr_test_db.tbl.id'])
                self.assertEqual(['attr_test_db.a3'], attrs['attr_test_db.tbl.s1'])
                self.assertEqual(['attr_test_db.a1'], attrs['attr_test_db.tbl.s1.f1'])
                self.assertEqual(['attr_test_db.a2'], attrs['attr_test_db.tbl.s1.s2'])
                self.assertEqual(['attr_test_db.a3'], attrs['attr_test_db.tbl.s1.s2.f2'])
                self.assertEqual(['attr_test_db.a1'], attrs['attr_test_db.tbl.s1.s2.f3'])

                describe_formatted = str(conn.execute_ddl(
                    'DESCRIBE FORMATTED %s.tbl' % db)).replace('\t', '  ')
                print(conn.execute_ddl_table_output('DESCRIBE %s.tbl' % db))
                print(describe_formatted)
                self.assertEqual(
                    expected_tbl_describe,
                    str(conn.execute_ddl_table_output('DESCRIBE %s.tbl' % db)))
                self.assertTrue(expected_tbl_describe_formatted in describe_formatted)

                if not cascade:
                    # If not cascade, create the view now to inherit.
                    conn.execute_ddl('''
                        CREATE VIEW %s.v1 AS SELECT
                          id, s1.f1 as f1, s1.s2.f2 as f2, s1.s2.f3 as f3
                        FROM %s.tbl''' % (db, db))
                    conn.execute_ddl('''
                        CREATE VIEW %s.v2 AS SELECT
                          s1, s1.s2 as s2
                        FROM %s.tbl''' % (db, db))
                    conn.execute_ddl('''
                        CREATE VIEW %s.v3 AS SELECT
                          id, s1.f1 as f, s1.s2 as s2
                        FROM %s.tbl''' % (db, db))
                    conn.execute_ddl('''
                        CREATE VIEW %s.v4 AS SELECT
                          id, s1.s2 as s2, s1.s2.f2 as f2
                        FROM %s.tbl''' % (db, db))

                print(conn.execute_ddl_table_output('DESCRIBE %s.v1' % db))
                self.assertEqual(
                    expected_v1_describe,
                    str(conn.execute_ddl_table_output('DESCRIBE %s.v1' % db)))

                print(conn.execute_ddl_table_output('DESCRIBE %s.v2' % db))
                self.assertEqual(
                    expected_v2_describe,
                    str(conn.execute_ddl_table_output('DESCRIBE %s.v2' % db)))

                print(conn.execute_ddl_table_output('DESCRIBE %s.v3' % db))
                self.assertEqual(
                    expected_v3_describe,
                    str(conn.execute_ddl_table_output('DESCRIBE %s.v3' % db)))

                print(conn.execute_ddl_table_output('DESCRIBE %s.v4' % db))
                self.assertEqual(
                    expected_v4_describe,
                    str(conn.execute_ddl_table_output('DESCRIBE %s.v4' % db)))

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.

#
# Tests for config CRUD APIs.
#
# pylint: disable=unused-argument

import unittest

from okera.tests import pycerebro_test_common as common
from okera._thrift_api import TConfigType, TRecordServiceException

upsert_config = common.upsert_config
list_configs = common.list_configs
delete_config = common.delete_config

def list_of_map_to_map(l, key_field):
    result = {}
    for v in l:
        key = v[key_field]
        result[key] = v
    return result

TEST_CONFIG_KEY1 = "test_configs_key1"
TEST_CONFIG_KEY2 = "test_configs_key2"

class ConfigsTest(unittest.TestCase):
    def test_basic(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Get and verify a system config
            configs = list_of_map_to_map(
                list_configs(conn, TConfigType.SYSTEM_CONFIG), 'key')
            self.assertTrue('audit.directory' in configs)
            self.assertEqual('/opt/okera/audit_logs', configs['audit.directory']['value'])

            # Upsert a new config and check the value
            # TODO: check return value
            upsert_config(conn, TConfigType.SYSTEM_CONFIG, [TEST_CONFIG_KEY1], "val1")
            configs = list_of_map_to_map(
                list_configs(conn, TConfigType.SYSTEM_CONFIG), 'key')
            self.assertEqual('val1', configs[TEST_CONFIG_KEY1]['value'])

            # Change the value
            upsert_config(conn, TConfigType.SYSTEM_CONFIG, [TEST_CONFIG_KEY1], "val2")
            configs = list_of_map_to_map(
                list_configs(conn, TConfigType.SYSTEM_CONFIG), 'key')
            self.assertEqual('val2', configs[TEST_CONFIG_KEY1]['value'])

            # Set to empty
            upsert_config(conn, TConfigType.SYSTEM_CONFIG, [TEST_CONFIG_KEY1], "")
            configs = list_of_map_to_map(
                list_configs(conn, TConfigType.SYSTEM_CONFIG), 'key')
            self.assertEqual('', configs[TEST_CONFIG_KEY1]['value'])

            # Add new config and check both
            upsert_config(conn, TConfigType.SYSTEM_CONFIG, [TEST_CONFIG_KEY2], "new_val1")
            configs = list_of_map_to_map(
                list_configs(conn, TConfigType.SYSTEM_CONFIG), 'key')
            self.assertEqual('', configs[TEST_CONFIG_KEY1]['value'])
            self.assertEqual('new_val1', configs[TEST_CONFIG_KEY2]['value'])

            # Delete first config
            delete_config(conn, TConfigType.SYSTEM_CONFIG, [TEST_CONFIG_KEY2])

            # Verify it is gone but other one is there
            configs = list_of_map_to_map(
                list_configs(conn, TConfigType.SYSTEM_CONFIG), 'key')
            self.assertEqual('', configs[TEST_CONFIG_KEY1]['value'])
            self.assertTrue(TEST_CONFIG_KEY2 not in  configs)

            # Delete the second config, verify both are gone
            delete_config(conn, TConfigType.SYSTEM_CONFIG, [TEST_CONFIG_KEY1])
            configs = list_of_map_to_map(
                list_configs(conn, TConfigType.SYSTEM_CONFIG), 'key')
            self.assertTrue(TEST_CONFIG_KEY1 not in  configs)
            self.assertTrue(TEST_CONFIG_KEY2 not in  configs)

    def test_abac(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            upsert_config(
                conn, TConfigType.SYSTEM_CONFIG, ['feature.ui.test_feature'], "true")
            upsert_config(
                conn, TConfigType.SYSTEM_CONFIG, ['test_feature'], "yes")
            query = 'SELECT * FROM okera_system.configs WHERE key = "test_feature"'
            superuser_list = conn.scan_as_json(query)
            anyuser_list = conn.scan_as_json(query, requesting_user='anyuser')

            self.assertEqual(1, len(superuser_list))
            self.assertEqual('test_feature', superuser_list[0].get('key'))
            self.assertEqual([], anyuser_list)

    def test_upsert_regex_autotagger(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Clean up tagging rules from  previous tests
            for c in list_configs(conn, TConfigType.AUTOTAGGER_REGEX):
                if c['namespace'] == 'test_ns' and c['tag'] == 'rule1':
                    delete_config(conn, TConfigType.AUTOTAGGER_REGEX, str(c['id']))

            params = {}
            params['name'] = 'great_rule'
            params['namespace'] = 'test_ns'
            params['description'] = 'a great regex'
            params['min_rows'] = '1'
            params['max_rows'] = '10'
            params['match_rate'] = '.4'
            params['match_column_name'] = 'true'
            params['match_column_comment'] = 'true'

            params['tag'] = 'rule1'
            params['expression'] = '*'
            upsert_config(conn, TConfigType.AUTOTAGGER_REGEX, None, params)

            rule = None
            for c in list_configs(conn, TConfigType.AUTOTAGGER_REGEX):
                if c['namespace'] == 'test_ns' and c['tag'] == 'rule1':
                    rule = c
                    break
            self.assertTrue(rule is not None)
            self.assertEqual('test_ns', rule['namespace'])
            self.assertEqual('*', rule['expression'])
            self.assertEqual('a great regex', rule['description'])
            self.assertTrue(rule['match_column_name'])
            self.assertTrue(rule['match_column_comment'])
            self.assertEqual('great_rule', rule['name'])

            # Update the expression
            params['expression'] = '.*'
            old_id = rule['id']
            upsert_config(conn, TConfigType.AUTOTAGGER_REGEX, [str(old_id)], params)

            rule = None
            for c in list_configs(conn, TConfigType.AUTOTAGGER_REGEX):
                if c['namespace'] == 'test_ns' and c['tag'] == 'rule1':
                    rule = c
                    break
            self.assertTrue(rule is not None)
            self.assertEqual('test_ns', rule['namespace'])
            self.assertEqual('.*', rule['expression'])
            self.assertEqual('great_rule', rule['name'])
            self.assertEqual(old_id, rule['id'])

            # Ensure we cannot update an expression removing the name
            del params['name']
            with self.assertRaises(TRecordServiceException):
                upsert_config(conn, TConfigType.AUTOTAGGER_REGEX, [str(old_id)], params)

            # Add another expression to the same tag
            params = {}
            params['name'] = 'greater_rule'
            params['namespace'] = 'test_ns'
            params['description'] = 'another great regex'
            params['min_rows'] = '3'
            params['max_rows'] = '55'
            params['match_rate'] = '.2'

            params['tag'] = 'rule1'
            params['expression'] = 'foo*'

            old_name = params['name']

            upsert_config(conn, TConfigType.AUTOTAGGER_REGEX, None, params)

            rules = []
            for c in list_configs(conn, TConfigType.AUTOTAGGER_REGEX):
                if c['namespace'] == 'test_ns' and c['tag'] == 'rule1':
                    rules.append(c)
            self.assertEqual(2, len(rules))
            self.assertTrue(old_id in (rules[0]['id'], rules[1]['id']))

            # Try to add a rule without a name
            params = {}
            params['namespace'] = 'test_ns'
            params['description'] = 'another great regex'
            params['min_rows'] = '3'
            params['max_rows'] = '55'
            params['match_rate'] = '.2'

            params['tag'] = 'rule1'
            params['expression'] = 'foo*'
            with self.assertRaises(TRecordServiceException):
                upsert_config(conn, TConfigType.AUTOTAGGER_REGEX, None, params)

            # Set the name to the empty string: this also should fail
            params['name'] = ''
            with self.assertRaises(TRecordServiceException):
                upsert_config(conn, TConfigType.AUTOTAGGER_REGEX, None, params)

            # Set the name to an existing name: this should fail
            params['name'] = old_name
            with self.assertRaises(TRecordServiceException):
                upsert_config(conn, TConfigType.AUTOTAGGER_REGEX, None, params)

if __name__ == "__main__":
    unittest.main()
